# (C) Copyright 2019 IBM Corp.
# (C) Copyright 2019 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Class to represent the concept and implementation of an association shrub.

A shrub is a view of association relations that gathers and presents all of the
information about a relation including the roles, reference classes, result
roles and result classes that may return instances. I differs from the
associator and reference commands in that they present the user with just a set
of CIM objects or their names, and not a view of the relations between the
components that make up a CIM association.

It is based on the parameters of the pywbem associators operation including
role, AssocClass, ResultRole, ResultClass. A shrub request that does not
include  these options, returns all possible relations of the association.
If any of the these optional parameters are included, only the names included
in the option are considered as the result is prepared.

It builds the information by using the reference and association operations to
gather data from the server and present either a table or tree view of
the relations that link a source instance and the target instances of an
association.
"""


from __future__ import absolute_import, print_function, unicode_literals

from collections import defaultdict, OrderedDict, namedtuple
import six
import click
# TODO: Future Could we combine this tree into tree file???
from asciitree import LeftAligned
from pydicti import dicti
from pywbem import CIMInstanceName, CIMClassName, CIMFloat, CIMInt, CIMError, \
    CIMDateTime

from ._common import output_format_is_table, format_table, shorten_path_str, \
    warning_msg
from ._utils import _ensure_unicode, _to_unicode

# Same as in pwbem.cimtypes.py
if six.PY2:
    # pylint: disable=invalid-name,undefined-variable
    _Longint = long  # noqa: F821
else:
    # pylint: disable=invalid-name
    _Longint = int

# Named tuple defining an Association class name and corresponding
# integer representing the Reference instance for that association. Used
# to display information about reference instances for associations with
# more than 2 reference properties.
AssocNameTuple = namedtuple('AssocName', 'Name, RefInst')


class AssociationShrub(object):
    # pylint: disable=useless-object-inheritance, too-many-instance-attributes
    """
    This class provides tools for the acquisition and display of an association
    that includes much more information than the DMTF defined operation
    Associatiors.  Using the same input parameters, it allows displaying
    the the components that make up an association as either a table or a
    tree including the reference classes, and roles.

    This class does not handle Error exceptions from the host WBEM Server.
    They must be handled by the user.
    """
    def __init__(self, context, source_path, Role=None, AssocClass=None,
                 ResultRole=None, ResultClass=None, fullpath=True,
                 verbose=None):
        self.source_path = source_path
        self.conn = context.conn
        self.context = context

        self.role = Role
        self.assoc_class = AssocClass
        self.result_role = ResultRole
        self.result_class = ResultClass

        self.fullpath = fullpath
        self.verbose = verbose

        #  Dictionary view of the shrub. This is a dictionary of dictionaries
        #  role:ReferenceClassNames:
        #  NOTE: OrderedDict is used to assure that table and tree outputs
        #  show results in deterministic order.
        self.instance_shrub = OrderedDict()

        # associated instance names dictionary organized by:
        #   - reference_class,
        #   - role,
        #   - associated class
        # NOTE: To account for issues where there is an error getting data from
        # host, the concept of a "None" role exists.
        self.assoc_instnames = OrderedDict()

        self.source_namespace = source_path.namespace or \
            self.conn.default_namespace

        self.source_host = source_path.host or self.conn.host

        # Cache for the results of conn.ReferenceNames(self.source+path)
        self.reference_names = None

        # Copy of source_path with namespace
        self.full_source_path = self.source_path.copy()
        if self.full_source_path.namespace is None:
            self.full_source_path.namespace = self.source_namespace

        self.ternary_ref_classes = OrderedDict()

        # Build the shrub dictionary for the instance defined by
        # self.source_path
        self._build_instance_shrub()

    @property
    def reference_classnames(self):
        """
        Return a list of the reference class names in the current shrub.
        This returns a list of objects of the class pywbem:CIMClassname
        which contains the name, host, and namespace for each class.
        """
        # pylint: disable=unnecessary-comprehension
        return [cln for cln in self.instance_shrub]

    def sorted_references(self, source_path):
        """
        Get reference instances from host sorted by instance name

        Parameters:

          source_path (:class:`pywbem:CIMInstancePath):
            Source path for the References request to the host

        Returns:
           list of :class:`pywbem:CIMClassNames` returns from host

        Raises:
            Error if Error returned from host
        """
        reference_instances = self.conn.References(source_path)
        return self.sort_instances(reference_instances)

    def sorted_associator_names(self, source_path, role=None, assoc_class=None,
                                result_role=None, result_class=None):
        """
        Get associated instances from host sorted by instance name.

        Parameters:

          source_path (:class:`pywbem:CIMInstancePath):

          role (:term:`string`):
            Optional string defining Role parameter of the Associators
            call to the host

          assoc_class (:class:`pywbem:CIMClass`):
            Optional string defining ResutRole parameter of the Associators
            call to the host.

          result_role (:term:`string`):
            Optional string defining ResutRole parameter of the Associators
            call to the host

          result_class (:class:`pywbem:CIMClass`):
            Optional string defining ResutRole parameter of the Associators
            call to the host

        Returns:
           list of :class:`pywbem:CIMClassNames` returns from host

        Raises:
            Error if error returned from host webserver
        """

        rtnd_assoc_inames = self.conn.AssociatorNames(
            source_path,
            Role=role,
            AssocClass=assoc_class,
            ResultRole=result_role,
            ResultClass=result_class)
        return self.sort_instance_names(rtnd_assoc_inames)

    def _build_instance_shrub(self):
        """
        Build the internal representation of a tree for the shrub as a
        dictionary of dictionaries representing the tree. This builds the shrub
        dictionary (self.instance_shrub) from the top of the tree down to
        the bottom.
        """
        # Build CIMClassname with host, namespace and insert if not
        # already in class_roles dictionary. Get the instance from
        # the host and roles from the instance.
        # ref_class_roles dictionary {<cln>:[roles]}
        reference_instances = self.sorted_references(self.source_path)

        self.define_ternary_references(reference_instances)

        # Build list of reference CIMClassName objects and add to a
        # ref_class_roles dict
        # Test for input --AssocClass parameter and limit ref names to this
        # class if it exists and is one of the defined classes
        reference_instnames = [i.path for i in reference_instances]

        if self.assoc_class:
            # Test if AssocClass parameter represents class in rtnd references
            if self.assoc_class.lower() in {n.classname.lower()
                                            for n in reference_instnames}:
                reference_instnames = [rn for rn in reference_instnames if
                                       self.assoc_class.lower() ==
                                       rn.classname.lower()]
            else:
                reference_instnames = [i.path for i in reference_instances]
                ref_clns = {n.classname for n in reference_instnames}
                warning_msg(
                    'Option --assoc-name "{}" not found in associator names '
                    '({})" from server'.format(self.assoc_class,
                                               ", ".join(ref_clns)))

        # Add the reference classes to the ref_class_roles dict
        ref_class_roles = OrderedDict()
        for ref in reference_instnames:
            cln = CIMClassName(ref.classname, ref.host, ref.namespace)
            if cln not in ref_class_roles:
                roles = self._get_reference_roles(ref)
                if roles:
                    ref_class_roles[cln] = roles

        # Find role parameter and insert into instance_shrub dictionary.
        # The result is dictionary of form:
        #   {<role>:{<ASSOC_CLASS>:[RESULTROLES]}
        for cln, roles in six.iteritems(ref_class_roles):
            role_dict = self._get_role_result_roles(roles, cln.classname)

            # Insert the role and cln into the shrub_dict
            for role, result_roles in role_dict.items():
                if role not in self.instance_shrub:
                    self.instance_shrub[role] = OrderedDict()

                # Put result_roles in shrub dict temporarily
                # Next code block converts value to
                if cln not in self.instance_shrub[role]:
                    self.instance_shrub[role][cln] = result_roles

        # If self.roles exists, remove any unwanted roles from dict
        # Accounts for case differences between self.role and roles from
        # target server
        if self.role:
            tst_role = self.role.lower()
            # Test for --role as valid role for this request and do warning.
            roles = list(self.instance_shrub)
            if tst_role not in [r.lower() for r in roles]:
                warning_msg('Option --role ({}) not found in roles: ({}). '
                            'Ignored'.format(self.role, ', '.join(roles)))
            else:
                # Remove any roles not defined in self.role
                remove_roles = [r for r in roles if r.lower() != tst_role]
                if remove_roles:
                    for role in remove_roles:
                        del self.instance_shrub[role]

        # Find associated instances/classes for each role
        for role in self.instance_shrub:
            self.assoc_instnames[role] = OrderedDict()
            for ref_classname in self.instance_shrub[role]:
                # Get temp installed result roles from shrub_dict and
                # replace with defaultdict that is basis for next level
                result_roles = self.instance_shrub[role][ref_classname]

                # Create associator result dictionaries with ref_class as key
                self.assoc_instnames[role][ref_classname] = OrderedDict()
                self.instance_shrub[role][ref_classname] = defaultdict(list)

                # Get Associated class names by AssocClass and ResultRole
                assoc_clns = []
                for result_role in result_roles:
                    # Get the associated instance names from server for
                    # all result_classes
                    if self.result_role:
                        if self.result_role.lower() != result_role.lower():
                            continue

                    rtnd_assoc_inames = self.sorted_associator_names(
                        self.source_path,
                        role=role,
                        assoc_class=ref_classname.classname,
                        result_role=result_role)

                    # Build unique associated classnames from returned inames.
                    # This is full ClassName entities including ns, host with
                    # a set comprehension so only unique names are kept
                    rtnd_classnames = list({CIMClassName(iname.classname,
                                                         iname.host,
                                                         iname.namespace)
                                            for iname in rtnd_assoc_inames})

                    # Discard unwanted assoc classes if --result_class param
                    # defined
                    if self.result_class:
                        rc = self.result_class.lower()
                        # Define list of unique returned assoc classnames
                        filtered_clns = list({iname.classname.lower() for iname
                                              in rtnd_assoc_inames
                                              if iname.classname.lower() == rc})

                        # Discard unwanted assoc classes
                        rtnd_classnames = [cln for cln in rtnd_classnames if
                                           cln.classname.lower()
                                           in filtered_clns]

                    assoc_clns.extend(rtnd_classnames)

                    # Extend instance_shrub_dict with returned assoc classnames
                    # pylint: disable=line-too-long
                    self.instance_shrub[role][ref_classname][result_role].extend(rtnd_classnames)  # noqa: E501
                    # pylint: enable=line-too-long

                    # Get associated instance names by AssocClass, role and
                    # target name using the assoc_clns from above
                    disp_result_role = result_role or "None"
                    # Get AssociatorNames for specific ResultClass
                    for assoc_cln in assoc_clns:
                        assoc_inames = self.sorted_associator_names(
                            self.source_path,
                            role=role,
                            assoc_class=ref_classname.classname,
                            result_role=result_role,
                            result_class=assoc_cln.classname)

                        # Build namedtuple of name, ref_inst  integer.
                        # This ties each output instance to a particular
                        # reference instance.
                        aname_tuples = OrderedDict()
                        for aname in assoc_inames:
                            for ref_inst_ctr, ref_inst in \
                                    enumerate(reference_instances):
                                if role not in ref_inst:
                                    continue
                                if ref_inst.get(role) != self.full_source_path:
                                    continue
                                # Find other properties with this result_role
                                # and create a tuple for each one found.
                                # The second data in the tuple identifies the
                                # reference instance by its position in the
                                # list of reference instances.
                                for name in ref_inst.properties:
                                    if name.lower() == result_role.lower():
                                        pvalue = ref_inst.properties[name].value
                                        anamex = aname.copy()
                                        anamex.host = None
                                        if pvalue == anamex:
                                            aname_tuples[aname] = \
                                                AssocNameTuple(
                                                    Name=aname,
                                                    RefInst=ref_inst_ctr)

                        # pylint: disable=line-too-long
                        if disp_result_role not in self.assoc_instnames[role][ref_classname]:  # noqa: E501
                            self.assoc_instnames[role][ref_classname][disp_result_role] = OrderedDict()  # noqa: E501
                        self.assoc_instnames[role][ref_classname][disp_result_role][assoc_cln] = list(aname_tuples.values())  # noqa: E501
                        # pylint: enable=line-too-long

    def display_shrub(self, output_format, summary=None):
        """
        Build the shrub output and display it to the output device based on
        the output_format.
        The default ouput format is ascii tree
        """

        if output_format_is_table(output_format):
            click.echo(self.build_shrub_table(output_format, summary))

        # default is display as ascii tree
        else:
            click.echo(self.build_ascii_display_tree(summary))

    def build_ascii_display_tree(self, summary):
        """
        Build ascii tree display for current shrub.
        Returns an String with the formatted ASCII tree
        """
        tree = self.build_shrub_tree(summary)

        tr = LeftAligned()
        return tr(tree)

    def build_shrub_tree(self, summary):
        """
        Prepare an ascii tree form of the shrub showing the hiearchy of
        components of the shrub. The top is the association source instance.
        The levels of the tree are:
            source instance
                role
                    reference_classe
                        result_role
                            result_classe
                                result_instances
        The dictionaries that define the tree are built from the bottom up
        based on the shrub_dict and use OrderedDict to preserve order
        of the items.
        """
        assoctree = OrderedDict()
        # Create dictionary of standard instance keys to potentially hide.
        # For now we always hide the following independent of key value
        replacements = dicti((("SystemCreationClassName", None),
                              ("SystemName", None)))
        for role, ref_clns in six.iteritems(self.instance_shrub):
            elementstree = OrderedDict()
            for ref_cln in ref_clns:
                rrole_dict = OrderedDict()
                for rrole, assoc_clns in six.iteritems(
                        self.assoc_instnames[role][ref_cln]):
                    assoc_clns_dict = OrderedDict()

                    for assoc_cln, inst_names in six.iteritems(assoc_clns):
                        if not inst_names:
                            continue

                        disp_assoc_cln = self.simplify_path(assoc_cln)
                        key = "{}(ResultClass)({} insts)". \
                            format(disp_assoc_cln, len(inst_names))

                        # Build dictionary of associated instance names
                        assoc_clns_dict[key] = OrderedDict()
                        if not summary:
                            # Insts dict is keys only with empty sub-dict
                            # for ascii tree compatibility. i.e. this
                            # is the lowest level in the tree.
                            # Returns OrderedDict
                            inst_names = self.build_inst_names(
                                inst_names,
                                ref_cln,
                                replacements,
                                self.fullpath)
                            assoc_clns_dict[key] = inst_names

                    # Add the role tree element
                    rrole_disp = "{}(ResultRole)".format(rrole)
                    rrole_dict[rrole_disp] = assoc_clns_dict

                # Add the reference class element. Include namespace if
                # different than conn default namespace
                disp_ref_cln = "{}(AssocClass)". \
                    format(self.simplify_path(ref_cln))

                elementstree[disp_ref_cln] = rrole_dict

            # Add the role component to the tree
            disp_role = "{}(Role)".format(role)
            assoctree[disp_role] = elementstree

        # Attach the top of the tree, the source instance path for the
        # shrub.
        display_source_path = self.simplify_path(self.source_path)
        toptree = {display_source_path: assoctree}

        return toptree

    def build_shrub_table(self, output_format, summary):
        """
        Build and return a table representing the shrub. The table
        returned is a string that can be printed to a terminal or or other
        destination.
        """
        def fmt_inst_col(iname_tuples, max_len, summary, ternary):
            """
            Format the instance column display either as a summary count
            or a list of instances possibly with attached integer representing
            reference instance and return it as a single string
            """
            if summary:
                return len(iname_tuples)

            if ternary:
                return "\n".join("{}(refinst:{})".format(
                    self.to_wbem_uri_folded(t[0], max_len=max_len), t[1])
                                 for t in iname_tuples)  # noqa E128

            return "\n".join("{}".format(
                self.to_wbem_uri_folded(t[0], max_len=max_len))
                             for t in iname_tuples)  # noqa E128

        # Display shrub as table
        inst_hdr = "Assoc Inst Count" if summary else "Assoc Inst paths"
        headers = ["Role", "AssocClass", "ResultRole", "ResultClass", inst_hdr]

        # Build the rows of the table
        rows = []
        # assoc_classnames dict struct [role]:[ref_clns]:[rrole]:[assoc_clns]
        for role, ref_clns in six.iteritems(self.instance_shrub):
            for ref_cln in ref_clns:
                is_ternary = self.ternary_ref_classes[ref_cln.classname]
                for rrole, assoc_clns in six.iteritems(
                        self.assoc_instnames[role][ref_cln]):
                    for assoc_cln in assoc_clns:
                        # pylint: disable=line-too-long
                        inst_names = self.assoc_instnames[role][ref_cln][rrole][assoc_cln]  # noqa E501
                        # pylint: enable=line-too-long
                        ml = click.get_terminal_size()[0] - 65
                        # TODO: ks: Create more general width algorithm
                        inst_col = fmt_inst_col(inst_names, ml, summary,
                                                is_ternary)

                        rows.append([role,
                                     self.simplify_path(ref_cln),
                                     rrole,
                                     self.simplify_path(assoc_cln),
                                     inst_col])

        title = 'Shrub of {}: {}'.format(self.source_path,
                                         'summary' if summary else 'paths')
        return format_table(rows, headers, title, table_format=output_format)

    def to_wbem_uri_folded(self, path, format='standard', max_len=15):
        # pylint: disable=redefined-builtin
        """
        Return the (untyped) WBEM URI string of this CIM instance path.
        This method modifies the pywbem:CIMInstanceName.to_wbem_uri method
        to return a slightly formated string where components are on
        separate lines if the length is longer than the max_len argument.

        See pywbem.CIMInstanceName.to_wbem_uri for detailed information

        Returns:

          :term:`unicode string`: Untyped WBEM URI of the CIM instance path,
          in the specified format.

        Raises:

          TypeError: Invalid type in keybindings
          ValueError: Invalid format
        """
        path = self.simplify_path(path)
        path_str = "{}".format(path)
        if len(path_str) <= max_len:
            return path_str

        ret = []

        def case(astring):
            """Return the string in the correct lexical case for the format."""
            if format == 'canonical':
                astring = astring.lower()
            return astring

        def case_sorted(keys):
            """Return the keys in the correct order for the format."""
            if format == 'canonical':
                case_keys = [case(k) for k in keys]
                keys = sorted(case_keys)
            return keys

        if format not in ('standard', 'canonical', 'cimobject', 'historical'):
            raise ValueError('Invalid format argument: {0}'.format(format))

        if path.host is not None and format != 'cimobject':
            # The CIMObject format assumes there is no host component
            ret.append('//')
            ret.append(case(path.host))

        if path.host is not None or format not in ('cimobject', 'historical'):
            ret.append('/')

        if path.namespace is not None:
            ret.append(case(path.namespace))

        if path.namespace is not None or format != 'historical':
            ret.append(':')

        ret.append(case(path.classname))

        ret.append('.\n')

        for key in case_sorted(path.keybindings.iterkeys()):
            value = path.keybindings[key]

            ret.append(key)
            ret.append('=')

            if isinstance(value, six.binary_type):
                value = _to_unicode(value)

            if isinstance(value, six.text_type):
                # string, char16
                ret.append('"')
                ret.append(value.
                           replace('\\', '\\\\').
                           replace('"', '\\"'))
                ret.append('"')
            elif isinstance(value, bool):
                # boolean
                # Note that in Python a bool is an int, so test for bool first
                ret.append(str(value).upper())
            elif isinstance(value, (CIMFloat, float)):
                # realNN
                # Since Python 2.7 and Python 3.1, repr() prints float numbers
                # with the shortest representation that does not change its
                # value. When needed, it shows up to 17 significant digits,
                # which is the precision needed to round-trip double precision
                # IEE-754 floating point numbers between decimal and binary
                # without loss.
                ret.append(repr(value))
            elif isinstance(value, (CIMInt, int, _Longint)):
                # intNN
                ret.append(str(value))
            elif isinstance(value, CIMInstanceName):
                # reference
                ret.append('"')
                ret.append(value.to_wbem_uri(format=format).
                           replace('\\', '\\\\').
                           replace('"', '\\"'))
                ret.append('"')
            elif isinstance(value, CIMDateTime):
                # datetime
                ret.append('"')
                ret.append(str(value))
                ret.append('"')
            else:
                raise TypeError(
                    "Invalid type {0} in keybinding value: {1}={2}"
                    .format(type(value), key, value))
            ret.append(',\n')

        del ret[-1]

        return _ensure_unicode(''.join(ret))

    def build_inst_names(self, inst_names_tuple, ref_cln, replacements,
                         fullpath=None):
        """
        Build a set of displayable instance names from the inst_names. This
        method tries to simplify the instance names by

        1. Hiding keys that have the same value for all instances. This
           is ignored if there is only a single instance
        2. Hiding certain specific key names that have a common meaning
        throughout the environment including SystemName,
        SystemCreationClassName, and CreationClassName.
        It hides CreationClassName key if the value is the same as the key
        classname.

        Next, if the defining reference class has more than 2 reference
        properties (ternary or greater associations) add an element to the
        instance name display indicate which reference instance is the
        connection for this association instance.

        Finally it removes the host and namespace if they are the same as
        the current host and namespace

        Parameters:

          inst_names_tuple (:func:`~py:collections.namedtuple` object):
            namedtuple containing instancename and integer representing
            reference instance.

          ref_cln (:term:`string`):
             Classname of the reference class.

          replacements (:class:`dicti`):
            No-case dictionary containing the name of each key to be considered
            for replacement with either the value None or a defined value
            for the key.  If the value is None the key will be replaced with
            '~' independent of its value.  If the value is not None, the
            key will be replaced only if value matches the key value.

          fullpath (boolean):
            If True, show the full instance paths.  If not True, build the path
            shortened by modifying selected keys to replace the keys defined
            by the replacements attribute with `~'. In addition all keys
            with the same value are replaced if len(inst_names) is gt 1.

        Returns:
            OrderedDict of modified key names
        """
        assert isinstance(inst_names_tuple, list)
        assert isinstance(inst_names_tuple[0], tuple)
        assert len(inst_names_tuple[0]) == 2
        assert isinstance(replacements, dicti)

        # If path shortening specified, determine which keys can be shortened
        # based on keys with the same value in all instance names
        if not fullpath:
            first_iname = inst_names_tuple[0][0]
            keys_to_hide = {k: True for k in first_iname.keys()}

            # Determine if there are multiple instances with same value
            if len(inst_names_tuple) > 1:
                for iname_tuple in inst_names_tuple:
                    iname = iname_tuple[0]
                    for kbname, kbvalue in iname.items():
                        if kbname not in replacements:
                            if kbvalue != first_iname.keybindings[kbname]:
                                keys_to_hide[kbname] = False
                replacements = {k: None for k, v in keys_to_hide.items() if v}

            # Test for CreationClassName. Hide if same as classname
            ccn = "creationclassname"
            for iname_tuple in inst_names_tuple:
                iname = iname_tuple[0]
                for kbname, kbvalue in iname.keybindings.items():
                    if kbname.lower() == ccn:
                        if iname.keybindings[ccn].lower() == \
                                iname.classname.lower():
                            replacements["CreationClassName"] = None

        ternary = self.ternary_ref_classes[ref_cln.classname]

        modified_inames = OrderedDict()
        for inst_name_tuple in inst_names_tuple:
            iname = self.simplify_path(inst_name_tuple.Name)
            # Convert path to string and possibly shorten
            iname_display = shorten_path_str(iname, replacements, fullpath)

            # If reference class with more than 2 references add indicator
            # to the defining reference instance so the user can match
            # the instances to reference instances.
            if ternary:
                iname_display = '{}(refinst:{})'.format(iname_display,
                                                        inst_name_tuple.RefInst)

            # builds dict with empty value to be ascii_tree compatible
            modified_inames[iname_display] = OrderedDict()
        return modified_inames

    def define_ternary_references(self, reference_instances):
        """
        Build dictionary of reference classes (ternary_ref_classes) in
        conn.References return with Value True if > 2 reference properties
        and False if == 2 reference properties
        """
        for ref_inst in reference_instances:
            count = 0
            if ref_inst.classname not in self.ternary_ref_classes:
                for v in ref_inst.properties.values():
                    if v.type == 'reference':
                        count += 1
                assert count >= 2  # Must return 2 or more reference properties
                # Set value True if count > 2 else False
                self.ternary_ref_classes[ref_inst.classname] = count > 2

    def simplify_path(self, path):
        """
        Simplify the CIMamespace instance defined by path by copying and
        removing the host name and namespace name if they are the same as
        the source instance.  This allows the tree to show only the
        classname for all components of the tree that are in the same
        namespace as the association source instance.
        """
        simple_path = path.copy()
        if simple_path.host and \
                simple_path.host.lower() == self.source_host.lower():
            simple_path.host = None
        if simple_path.namespace and \
                simple_path.namespace.lower() == self.source_namespace.lower():
            simple_path.namespace = None
        return simple_path

    def _get_reference_roles(self, inst_name):
        """
        Internal method to get the list of roles for an association class.
        Uses instance get rather than class get because some
        servers may not support class get operation.
        """
        try:
            ref_inst = self.conn.GetInstance(inst_name, LocalOnly=False)
        except CIMError as ce:
            click.echo('Exception ref {0}, exception: {1}'.format(inst_name,
                                                                  ce))
            return None
        roles = [pname for pname, pvalue in six.iteritems(ref_inst.properties)
                 if pvalue.type == 'reference']
        if self.verbose:
            print('class {0}, roles {1}'.format(inst_name.classname, roles))
        return roles

    def _get_role_result_roles(self, roles, ref_classname):
        """
        Given the reference classname, separate the role and result_role
        parameters and return them. This method determines that the role
        is the call to ReferenceNames that returns references. Result roles
        are the roles that do not return references. Note that there are
        cases where this basic algorithm returns multiples
        """
        rtn_roles = OrderedDict()
        for tst_role in roles:
            refs = self.conn.ReferenceNames(self.source_path,
                                            Role=tst_role,
                                            ResultClass=ref_classname)
            self.reference_names = refs

            if refs:
                rtn_roles[tst_role] = [r for r in roles if r != tst_role]
        if self.verbose:
            print('ResultRoles: class={0} ResultClass={1} ResultRoles={2}'
                  .format(self.source_path.classname, ref_classname, rtn_roles))
        return rtn_roles

    @staticmethod
    def sort_instances(instances):
        """
        Sort list of CIMInstanceName objects.  To sort an instance name
        it must be converted to a string where gt and lt compares can be
        executed.  to_wbem_uri(formt='canonical') accomplishes that

        Parameters:

          inames (list of CIMInstanceName):

        Returns:
            list of CIMInstance name sorted.
        """
        if len(instances) <= 1:
            return instances

        d = {}
        for instance in instances:
            key = instance.path.to_wbem_uri(format="canonical")
            d[key] = instance

        return [d[k] for k in sorted(d.keys())]

    @staticmethod
    def sort_instance_names(inames):
        """
        Sort list of CIMInstanceName objects.  To sort an instance name
        it must be converted to a string where gt and lt compares can be
        executed.  to_wbem_uri(formt='canonical') accomplishes that

        Parameters:

          inames (list of CIMInstanceName):

        Returns:
            list of CIMInstance name sorted.
        """
        if len(inames) <= 1:
            return inames

        d = {}
        for iname in inames:
            key = iname.to_wbem_uri(format="canonical")
            d[key] = iname

        return [d[k] for k in sorted(d.keys())]



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Pywbemcli function to format CIM values from properties and parameters into
strings that have defined folding rules and can fit into specific spaces like
table cells
"""

from __future__ import absolute_import, print_function, unicode_literals

import six

from pywbem import CIMInstance, CIMClass, CIMFloat, CIMInt

# Same as in pywbem.cimobj.py
try:
    from builtins import type as builtin_type
except ImportError:  # py2
    from __builtin__ import type as builtin_type


DEFAULT_MAX_CELL_WIDTH = 100

# Same as in pwbem.cimtypes.py
if six.PY2:
    # pylint: disable=invalid-name,undefined-variable
    _Longint = long  # noqa: F821
else:
    # pylint: disable=invalid-name
    _Longint = int


# Constants for string formatting
DEFAULT_INDENT = 3

#
#  NOTE: The following code was taken largely from pywbem methods that
#  prepare strings for MOF output. We kept the same basic formatting but
#  simplified slightly.
#  The goal is to create strings for property or parameter values that can be
#  placed in rows of a table and that follow the detailed formatting of MOF
#
#  The names were changed in some cases and all of the functions except
#  the function cimvalue_to_fmtd_string(...) are marked internal.
#
#
#  We have left most of the options in that are in pywbem because we are
#  unsure how this will be used and that can affect formatting in the future.
#


def _indent_str(indent):
    """
    Return a MOF indent pad unicode string from the indent integer variable
    that defines number of spaces to indent. Used to format MOF output.
    """
    return u''.ljust(indent, u' ')


def _mofval(value, indent, maxline, line_pos=0, end_space=0):
    """
    Low level function that returns the MOF representation of a non-string
    value (i.e. a value that cannot not be split into multiple parts, for
    example a numeric or boolean value).

    If the MOF representation of the value does not fit into the remaining
    space of the current line, it is put into a new line, considering the
    specified indentation.

    NOTE: This method is derived from pywbem mofval but differs in that we
    want to output even if we violate the maxline limit on the new line. This
    method favors outputting data over exceptions.

    Parameters:

      value (:term:`unicode string`): The non-string value. Must not be `None`.

      indent (:term:`integer`): Number of spaces to indent any new lines that
        are generated.

      maxline (:term:`integer`): Maximum line length for the generated MOF.

      line_pos (:term:`integer`): Length of content already on the current
        line.

      end_space (:term:`integer`): Length of space to be left free on the last
        line.

    Returns:

      tuple of
        * :term:`unicode string`: MOF string.
        * new line_pos
    """

    assert isinstance(value, six.text_type)

    # Check for output on current line
    # if fits or this is first entry on the line
    avl_len = maxline - line_pos - end_space
    if len(value) <= avl_len or line_pos == 0:
        line_pos += len(value)
        return value, line_pos

    mof_str = u'\n' + _indent_str(indent) + value
    line_pos = indent + len(value)
    return mof_str, line_pos


def _scalar_value_tomof(value, type, indent, maxline, line_pos=0, end_space=0,
                        avoid_splits=False):
    # pylint: disable=line-too-long,redefined-builtin
    """
    Return a MOF string representing a scalar CIM-typed value.

    `None` is returned as 'NULL'.

    NOTE: This code taken from pywbem 0.14.0

    Parameters:

      value (:term:`CIM data type`, :term:`number`, :class:`~pywbem.CIMInstance`, :class:`~pywbem.CIMClass`):
        The scalar CIM-typed value. May be `None`.

        Must not be an array/list/tuple. Must not be a :ref:`CIM object` other
        than those listed.

      type (string): CIM data type name.

      indent (:term:`integer`): Number of spaces to indent any new lines that
        are generated.

      maxline (:term:`integer`): Maximum line length for the generated MOF.

      line_pos (:term:`integer`): Length of content already on the current
        line.

      end_space (:term:`integer`): Length of space to be left free on the last
        line.

      avoid_splits (bool): Avoid splits at the price of starting a new line
        instead of using the current line.

    Returns:

      tuple of
        * :term:`unicode string`: MOF string.
        * new line_pos
    """  # noqa: E501

    if type == 'string':  # pylint: disable=no-else-raise
        if isinstance(value, six.string_types):
            return _mofstr(value, indent, maxline, line_pos, end_space,
                           avoid_splits)

        if isinstance(value, (CIMInstance, CIMClass)):
            # embedded instance or class
            return _mofstr(value.tomof(), indent, maxline, line_pos, end_space,
                           avoid_splits)
        raise TypeError("Scalar value of CIM type {0} has invalid Python type "
                        "type {1} for conversion to a MOF string".format
                        (type, builtin_type(value)))

    elif type == 'char16':
        return _mofstr(value, indent, maxline, line_pos, end_space,
                       avoid_splits, quote_char=u"'")
    elif type == 'boolean':
        val = u'true' if value else u'false'
        return _mofval(val, indent, maxline, line_pos, end_space)
    elif type == 'datetime':
        val = six.text_type(value)
        return _mofstr(val, indent, maxline, line_pos, end_space, avoid_splits)
    elif type == 'reference':
        val = value.to_wbem_uri(format='standard')
        return _mofstr(val, indent, maxline, line_pos, end_space, avoid_splits)
    elif isinstance(value, (CIMFloat, CIMInt, int, _Longint)):
        val = six.text_type(value)
        return _mofval(val, indent, maxline, line_pos, end_space)
    else:
        assert isinstance(value, float), \
            "Scalar value of CIM type {0} has invalid Python type {1} " \
            "for conversion to a MOF string".format(type, builtin_type(value))
        val = repr(value)
        return _mofval(val, indent, maxline, line_pos, end_space)


def mof_escaped(strvalue):
    # Note: This is a raw docstring because it shows many backslashes, and
    # that avoids having to double them.
    r"""
    Return a MOF-escaped string from the input string.

    Parameters:

      strvalue (:term:`unicode string`): The string value. Must not be `None`.
        Special characters must not be backslash-escaped.

    Details on backslash-escaping:

    `DSP0004` defines that the character repertoire for MOF string constants
    is the entire repertoire for the CIM string datatype. That is, the entire
    Unicode character repertoire except for U+0000.

    The only character for which `DSP0004` requires the use of a MOF escape
    sequence in a MOF string constant, is the double quote (because a MOF
    string constant is enclosed in double quotes).

    `DSP0004` defines MOF escape sequences for several more characters, but it
    does not require their use in MOF. For example, it is valid for a MOF
    string constant to contain the (unescaped) characters U+000D (newline) or
    U+0009 (horizontal tab), and others.

    Processing the MOF escape sequences as unescaped characters may not be
    supported by MOF-related tools, and therefore this function plays it safe
    and uses the MOF escape sequences defined in `DSP0004` as much as possible.
    The following table shows the MOF escape sequences defined in `DSP0004`
    and whether they are used (i.e. generated) by this function:

    ========== ==== ===========================================================
    MOF escape Used Character
    sequence
    ========== ==== ===========================================================
    \b         yes  U+0008: Backspace
    \t         yes  U+0009: Horizontal tab
    \n         yes  U+000A: Line feed
    \f         yes  U+000C: Form feed
    \r         yes  U+000D: Carriage return
    \"         yes  U+0022: Double quote (") (required to be used)
    \'         yes  U+0027: Single quote (')
    \\         yes  U+005C: Backslash (\)
    \x<hex>    (1)  U+<hex>: Any UCS-2 character, where <hex> is one to four
                      hex digits, representing its UCS code position (this form
                      is limited to the UCS-2 character repertoire)
    \X<hex>    no   U+<hex>: Any UCS-2 character, where <hex> is one to four
                      hex digits, representing its UCS code position (this form
                      is limited to the UCS-2 character repertoire)
    ========== ==== ===========================================================

    (1) Yes, for all other characters in the so called "control range"
        U+0001..U+001F.
    """

    escaped_str = strvalue

    # Escape backslash (\)
    escaped_str = escaped_str.replace('\\', '\\\\')

    # Escape \b, \t, \n, \f, \r
    # Note, the Python escape sequences happen to be the same as in MOF
    escaped_str = escaped_str.\
        replace('\b', '\\b').\
        replace('\t', '\\t').\
        replace('\n', '\\n').\
        replace('\f', '\\f').\
        replace('\r', '\\r')

    # Escape remaining control characters (U+0001...U+001F), skipping
    # U+0008, U+0009, U+000A, U+000C, U+000D that are already handled.
    # We hard code it to be faster, plus we can easily skip already handled
    # chars.
    # The generic code would be (not skipping already handled chars):
    #     for cp in range(1, 32):
    #         c = six.unichr(cp)
    #         esc = '\\x{0:04X}'.format(cp)
    #         escaped_str = escaped_str.replace(c, esc)
    escaped_str = escaped_str.\
        replace(u'\u0001', '\\x0001').\
        replace(u'\u0002', '\\x0002').\
        replace(u'\u0003', '\\x0003').\
        replace(u'\u0004', '\\x0004').\
        replace(u'\u0005', '\\x0005').\
        replace(u'\u0006', '\\x0006').\
        replace(u'\u0007', '\\x0007').\
        replace(u'\u000B', '\\x000B').\
        replace(u'\u000E', '\\x000E').\
        replace(u'\u000F', '\\x000F').\
        replace(u'\u0010', '\\x0010').\
        replace(u'\u0011', '\\x0011').\
        replace(u'\u0012', '\\x0012').\
        replace(u'\u0013', '\\x0013').\
        replace(u'\u0014', '\\x0014').\
        replace(u'\u0015', '\\x0015').\
        replace(u'\u0016', '\\x0016').\
        replace(u'\u0017', '\\x0017').\
        replace(u'\u0018', '\\x0018').\
        replace(u'\u0019', '\\x0019').\
        replace(u'\u001A', '\\x001A').\
        replace(u'\u001B', '\\x001B').\
        replace(u'\u001C', '\\x001C').\
        replace(u'\u001D', '\\x001D').\
        replace(u'\u001E', '\\x001E').\
        replace(u'\u001F', '\\x001F')

    # Escape single and double quote
    escaped_str = escaped_str.replace('"', '\\"')
    escaped_str = escaped_str.replace("'", "\\'")

    return escaped_str


def _mofstr(value, indent, maxline, line_pos, end_space, avoid_splits=False,
            quote_char=u'"'):
    """
    Low level function that returns the MOF representation of a string value
    (i.e. a value that can be split into multiple parts, for example a string,
    reference or datetime typed value).

    The function performs the backslash-escaping of characters in the string
    (for details, see function mof_escaped()), handles the splitting into
    multiple string parts if the current line does not have sufficient space
    left, and surrounds the string parts (or the entire string, if it ends up
    having only one part) with the specified quote characters.

    The strategy for starting new lines and for splitting the string into parts
    is:

    * If the string fits into the current line, it is output.
    * If the 'avoid_splits' flag is set, a new line is generating. If the
      string fits onto the new line, it is output. Otherwise, the string is
      split into parts and these are output starting with the new line,
      generating additional new lines as needed.
    * If the 'avoid_splits' flag is not set, the string is split into parts and
      these are output starting with the current line, generating new lines as
      needed.
    * Strings are first tried to split after the rightmost space character that
      would still make it fit onto the line, and only if there is no space
      character in that range, the string is split at a non-space position.

    Parameters:

      value (:term:`unicode string`): The string value. Must not be `None`.
        Special characters must not be backslash-escaped.

      indent (:term:`integer`): Number of spaces to indent any new lines that
        are generated.

      maxline (:term:`integer`): Maximum line length for the generated string.

      line_pos (:term:`integer`): Length of content already on the current
        line.

      end_space (:term:`integer`): Length of space to be left free on the last
        line.

      avoid_splits (bool): Avoid splits at the price of starting a new line
        instead of using the current line.

      quote_char (:term:`unicode string`): Character to be used for surrounding
        the string parts with. For CIM string typed values, this must be a
        double quote (the default), and for CIM char16 typed values, this must
        be a single quote.

    Returns:

      tuple of
        * :term:`unicode string`: MOF string.
        * new line_pos
    """

    assert isinstance(value, six.text_type)

    value = mof_escaped(value)

    quote_len = 2  # length of the quotes surrounding a string part
    new_line = u'\n' + _indent_str(indent)

    mof = []

    while True:

        # Prepare safety check for endless loops
        saved_value = value

        avl_len = maxline - line_pos - quote_len

        # Decide whether to start a new line
        if len(value) > avl_len - end_space:
            if avoid_splits or avl_len < 0:
                # Start a new line
                mof.append(new_line)
                line_pos = indent
                avl_len = maxline - indent - quote_len
            else:
                # Find last fitting blank
                blank_pos = value.rfind(u' ', 0, avl_len)
                if blank_pos < 0:
                    # We cannot split at a blank -> start a new line
                    mof.append(new_line)
                    line_pos = indent
                    avl_len = maxline - indent - quote_len

        # Check whether the entire string fits (that is a last line, then)
        if len(value) <= avl_len - end_space:
            mof.append(quote_char)
            mof.append(value)
            mof.append(quote_char)
            line_pos += quote_len + len(value)
            break

        # Split the string and output the next part
        split_pos = value.rfind(u' ', 0, avl_len)
        if split_pos < 0:
            # We have to split within a word
            split_pos = avl_len - 1
        part_value = value[0:split_pos + 1]
        value = value[split_pos + 1:]
        mof.append(quote_char)
        mof.append(part_value)
        mof.append(quote_char)
        line_pos += quote_len + len(part_value)

        if value == u'':
            break

        # A safety check for endless loops
        assert value != saved_value, \
            "Endless loop in _mofstr() with state: " \
            "mof_str={0}, value={1}, avl_len={2}, end_space={3}, " \
            "split_pos={4}". \
            format(u''.join(mof), value, avl_len, end_space, split_pos)

    mof_str = u''.join(mof)
    return mof_str, line_pos


def cimvalue_to_fmtd_string(value, type, indent=0,
                            maxline=DEFAULT_MAX_CELL_WIDTH,
                            line_pos=0, end_space=0, avoid_splits=False,
                            valuemapping=None):
    # pylint: disable=redefined-builtin
    """
    Return a MOF string representing a CIM-typed value (scalar or array).

    In case of an array, the array items are separated by comma, but the
    surrounding curly braces are not added.

    In case a value mapping is provided, each value is followed by the
    Values qualifier string in parenthesis. For example, "4" may become
    "4 (Enabled)".

    Parameters:

      value (CIM-typed value or list of CIM-typed values): The value.

      indent (:term:`integer`): Number of spaces to indent any new lines that
        are generated.

      maxline (:term:`integer`): Maximum line length for the generated MOF.

      line_pos (:term:`integer`): Length of content already on the current
        line.

      end_space (:term:`integer`): Length of space to be left free on the last
        line.

      avoid_splits (bool): Avoid splits at the price of starting a new line
        instead of using the current line.

      valuemapping (:class:`pywbem.ValueMapping`): None or a value mapping
        defining a string for the integer-typed property value(s).

    Returns:

      tuple of
        * :term:`unicode string`: MOF string.
        * new line_pos
    """

    if isinstance(value, list):

        mof = []

        for i, v in enumerate(value):

            if i > 0:
                # Assume comma and space as separator
                line_pos += 2

            val_str, line_pos = _scalar_value_tomof(
                v, type, indent, maxline, line_pos, end_space + 2, avoid_splits)
            if valuemapping:
                val_str = "{} ({})".format(val_str, valuemapping.tovalues(v))

            if i > 0:
                # Add the actual separator
                mof.append(u',')
                if val_str[0] != '\n':
                    mof.append(u' ')
                else:
                    # Adjust by the space we did not need
                    line_pos -= 1

            mof.append(val_str)

        mof_str = u''.join(mof)

    else:
        mof_str, line_pos = _scalar_value_tomof(
            value, type, indent, maxline, line_pos, end_space, avoid_splits)
        if valuemapping:
            mof_str = "{} ({})".format(mof_str, valuemapping.tovalues(value))
    return mof_str, line_pos



#!/usr/bin/env python
"""
This script removes duplicated dist-info directories of setuptools, which
happen to exist on the Travis CI in their Ubuntu 14.04 (trusty) distro:
  site-packages/setuptools                  <- contains 36.3.0
  site-packages/setuptools-36.0.1.dist-info <- duplicate to be removed
  site-packages/setuptools-36.3.0.dist-info
"""

import sys
import os
import re
import glob
import importlib
import shutil


def remove_duplicate_metadata_dirs(package_name):
    """Remove duplicate metadata directories of a package."""

    print("Removing duplicate metadata directories of package: %s" %
          package_name)

    module = importlib.import_module(package_name)

    py_mn = "{0}.{1}".format(sys.version_info[0], sys.version_info[1])
    print("Current Python version: {0}".format(py_mn))

    version = module.__version__
    print("Version of imported {0} package: {1}".format(package_name, version))

    site_dir = os.path.dirname(os.path.dirname(module.__file__))
    print("Site packages directory of imported package: {0}".format(site_dir))

    metadata_dirs = []
    metadata_dirs.extend(glob.glob(os.path.join(
        site_dir, '{0}-*.dist-info'.format(package_name))))
    metadata_dirs.extend(glob.glob(os.path.join(
        site_dir, '{0}-*-py{1}.egg-info'.format(package_name, py_mn))))

    for d in metadata_dirs:

        m = re.search(r'/%s-([0-9.]+)(\.di|-py)' % package_name, d)

        if not m:
            print("Warning: Could not parse metadata directory: {0}".format(d))
            continue

        d_version = m.group(1)

        if d_version == version:
            print("Found matching metadata directory: {0}".format(d))
            continue

        print("Removing duplicate metadata directory: {0}".format(d))
        shutil.rmtree(d)


if __name__ == '__main__':
    remove_duplicate_metadata_dirs('setuptools')



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Common Functions applicable across multiple components of pywbemcli
"""

from __future__ import absolute_import, print_function, unicode_literals

import fnmatch
import re
from textwrap import fill
from operator import itemgetter
from collections import namedtuple
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict  # pylint: disable=import-error

from pydicti import odicti
import six
import click
import tabulate

from pywbem import CIMInstanceName, CIMInstance, CIMClass, \
    CIMQualifierDeclaration, CIMProperty, CIMClassName, \
    cimvalue, ValueMapping

from .config import USE_TERMINAL_WIDTH, DEFAULT_TABLE_WIDTH

from ._cimvalueformatter import cimvalue_to_fmtd_string


# Definitions of the components of the help Usage line used
# by pywbemcli
GENERAL_OPTS_TXT = '[GENERAL-OPTIONS]'
CMD_OPTS_TXT = '[COMMAND-OPTIONS]'
SUBCMD_HELP_TXT = "COMMAND [ARGS] " + CMD_OPTS_TXT

DEFAULT_MAX_CELL_WIDTH = 100

INT_TYPE_PATTERN = re.compile(r'^[su]int(8|16|32|64)$')

##############################################################
#
#     General option output format definitions and support for
#     validating output formats
#
##############################################################

# Definition of the format keywords within each format group
TABLE_FORMATS = ('table', 'plain', 'simple', 'grid', 'psql', 'rst', 'html')
CIM_OBJECT_FORMATS = ('mof', 'xml', 'repr', 'txt')
TEXT_FORMATS = ('text',)

# Definition of the default format for each format group
DEFAULT_CIM_FORMAT = CIM_OBJECT_FORMATS[0]
DEFAULT_TABLE_FORMAT = TABLE_FORMATS[2]
DEFAULT_TEXT_FORMAT = TEXT_FORMATS[0]

# Dictionary that relates group names to list of formats keywords for that
# group and the default format for that group.  The order of the dictionary
# entry determines the priority of selecting formats.
GROUPVALUES = namedtuple('GROUPVALUES', 'keywords default')
OUTPUT_FORMAT_GROUPS = OrderedDict(
    [('CIM', GROUPVALUES(CIM_OBJECT_FORMATS, DEFAULT_CIM_FORMAT)),
     ('TABLE', GROUPVALUES(TABLE_FORMATS, DEFAULT_TABLE_FORMAT)),
     ('TEXT', GROUPVALUES(TEXT_FORMATS, DEFAULT_TEXT_FORMAT))])

# All of the format strings
OUTPUT_FORMATS = tuple(item for fmts in OUTPUT_FORMAT_GROUPS.values()
                       for item in fmts.keywords)


def output_format_is_table(output_format):
    """
    Return True if output format is a table form.

    Parameters:

      output_format (:term:`string`):
         String containing the output_format keyvalue.

    Returns:
      True: if `output_format` is one of the Table format keyword choises
      False: if `output_format` is not a Table format.
    """
    return output_format in TABLE_FORMATS


def validate_output_format(output_format, valid_format_groups,
                           default_format=None):
    """
    Tests for valid format groups and provides a default format if the
    context.output_format is None.

    Parameters:

      output_format (:term:`string` or None):
        The output format string to be validated (normally provided by input
        (i.e. ContextObj)) or None if there is no defined format for this
        command execution.

      valid_format_groups (list of :term:`string` or :term:`string`):
        One or more strings in a list where the allowed strings are: 'TABLE',
        'CIM', or 'TEXT'. An empty list implies any output group if valid. A
        single string may be used to designate a single group

      default_format (:term:`string` or None):
        One of the valid format definitions or None.  The format must be in the
        OUTPUT_FORMATS list. If None, the default format for the first group in
        `valid_format_groups` is returned. This format keyword is returned if
        `output_format` is None.

    Returns:
       :term:`string` containing output format keyword to be used.

    Raises:
        click.ClickException if format invalid for command
    """

    # These are asserts because they are coding errors in the input parameters
    # for this function
    if isinstance(valid_format_groups, six.string_types):
        valid_format_groups = [valid_format_groups]

    # Code error if this assert fails
    assert all(element in OUTPUT_FORMAT_GROUPS for element in
               valid_format_groups)

    if default_format:
        assert default_format in OUTPUT_FORMATS

    # Confirm default matches group.
    # CIM object type has priority over Table
    if output_format:
        # Do invalid message here but really should be assert because this
        # is a error in the data defined for the validate call.
        for group, value in OUTPUT_FORMAT_GROUPS.items():
            if group in valid_format_groups:
                if output_format in value.keywords:
                    return output_format
        if not valid_format_groups:
            return output_format

    else:   # output_format is None
        if default_format:
            return default_format
        if valid_format_groups:
            for group, value in OUTPUT_FORMAT_GROUPS.items():
                if valid_format_groups[0] == group:
                    return value.default
            return OUTPUT_FORMAT_GROUPS['CIM'].default
        return OUTPUT_FORMAT_GROUPS['CIM'].default

    valid_formats = ""
    for name, value in OUTPUT_FORMAT_GROUPS.items():
        fmt_list = value.keywords
        if name in valid_format_groups:
            if valid_formats:
                valid_formats += "; "
            valid_formats += '{0} formats: "({1})"'.format(
                name, '", "'.join(fmt_list))

    raise click.ClickException('Output format "{}" not allowed for this '
                               'command. Only {} allowed.'.
                               format(output_format, valid_formats))


######################################################################
#
#  General common functions
#
######################################################################


def resolve_propertylist(propertylist):
    """
    Resolve property list received from click options.  Click options produces
    an empty list when there is no property list.

    Pywbem requires None when there is no propertylist

    Further, property lists can be input as a comma separated list so this
    function also splits any string with embedded commas.

    Parameters:

      propertylist (list of :term:`string` or None):
        Each item in list may be a single property name or a collection of
        property names separated by commas.

    Returns:
        list of property names resulting from parsing input or empty list
        or None
    """
    # If no property list, return None which means all properties
    if not propertylist:
        return None

    # If propertylist is a single empty string, set to empty list.
    if len(propertylist) == 1 and not propertylist[0]:
        propertylist = []

    # expand any comma separated entries in the list
    else:
        pl = []
        for item in propertylist:
            if ',' in item:
                pl.extend(item.split(','))
            else:
                pl.append(item)
        propertylist = pl

    return propertylist


def warning_msg(msg):
    """Issue the msg param as warning prefixed by WARNING: to stderr"""
    click.echo('WARNING: {}'.format(msg), err=True)


######################################################################
#
#  Functions to select from console
#
######################################################################


def pick_one_from_list(context, options, title):
    """
    Interactive component that displays a set of options (strings) and asks
    the user to select one.  Returns the item and index of the selected string.

    If there is only a single item in the options, simply return that choice
    without user intervention.

    Parameters:
      options:
        List of strings from which one will is to be selected

      title:
        Title to display before selection

    Retries until either integer within range of options list is input
    or user enter no value. Ctrl-C ends even the REPL.

    Returns:
      Selected item from options_list

    Raises:
      ValueError if Ctrl-c input from console.

    TODO: Possible Future This could be replaced by the python pick library
    that would use curses for the selection process.
    """

    # If there is only a single choice, return that choice.
    if len(options) == 1:
        return options[0]

    # Issue list of choices and prompt for user choice of index
    if context:
        context.spinner_stop()

    click.echo(title)
    index = -1
    for str_ in options:
        index += 1
        click.echo('{}: {}'.format(index, str_))
    selection = None
    msg = 'Input integer between 0 and {} or Ctrl-C to exit selection' \
        .format(index)

    # Loop for valid user choice until valid choice made or selection aborted
    # by user
    while True:
        try:
            selection_txt = click.prompt(msg)
            selection = int(selection_txt)
            if 0 <= selection <= index:
                if context:
                    context.spinner_start()
                return options[selection]
        except ValueError:  # This causes the retry of the request
            pass
        except KeyboardInterrupt:
            raise click.ClickException("Pick Aborted. CTRL-C")
        except Exception as ex:
            raise click.ClickException(
                'Selection exception: {} Command Aborted'.format(ex))
        click.echo('"{}" Invalid response {}'.format(selection_txt, msg))


def pick_instance(context, objectname, namespace=None):
    """
    Display list of instances names from provided classname to console and user
    selects one. Returns the selected instancename.

    Parameters:

      context:
        Current click context or None

      classname:
        Classname to use to get instance names from server

    Returns:
        instancename selected or None if there are no instances to pick

    Raises:
        ClickException if user choses to terminate the selection process
    """
    if not is_classname(objectname):
        raise click.ClickException('{} must be a classname'.format(objectname))
    instance_names = context.conn.PyWbemcliEnumerateInstancePaths(objectname,
                                                                  namespace)

    if not instance_names:
        click.echo('No instance paths found for {}'.format(objectname))
        return None

    try:
        return pick_one_from_list(context, instance_names,
                                  'Pick Instance name to process')
    except Exception as ex:
        raise click.ClickException('Command Aborted. Exception {}'.format(ex))


def pick_multiple_from_list(context, options, title):
    """
    Interactive component that displays a set of options (strings) and asks
    the user to select multiple entries from that list.  Returns a list of
    the items selected.

    Retries until either integer within range of options list is input
    or user enter no value. Ctrl-C ends even the REPL.

    Parameters:
      context:
        If not None, the ContextObj which is used to stop and start the
        spinner.
      options:
        List of strings to select

      title:
        Title to display before selection

    Returns:
      list of index of selected items

    Raises:
      ValueError if Ctrl-c input from console.

    TODO: This could be replaced by the python pick library that would use
    curses for the selection process.
    """
    if context:
        context.spinner_stop()

    click.echo(title)
    index = -1
    for str_ in options:
        index += 1
        click.echo('{}: {}'.format(index, str_))
    selection = None
    selection_list = []
    msg = 'Select entry by index or hit enter to end selection>'
    while True:
        try:
            selection_txt = click.prompt(msg)
            if not selection_txt:
                if context:
                    context.spinner_start()
                return selection_list

            selection = int(selection_txt)
            if 0 <= selection <= index:
                selection_list.append(options[selection])
            continue
        except ValueError:
            pass
        except KeyboardInterrupt:
            raise ValueError
        click.echo('{} Invalid. Input integer between 0 and {}; hit enter to '
                   'stop selection.'.format(selection, index))


def is_classname(astring):
    """
    Test if the astring input is a classname or contains instance name
    components.  The existence of a period at the end of the name component
    determines if it is a classname or instance name.

    Returns:
        True if classname. Otherwise it returns False
    """
    assert isinstance(astring, six.string_types)
    return not re.match(r'[a-zA_Z0-9_].*\.', astring)


def filter_namelist(pattern, name_list, ignore_case=True):
    """
    Filter out names in name_list that do not match glob pattern compiled
    as regex.

    The regex is defined as IGNORECASE and anchored.

    Note that the regex may define a subset of the name string.  Thus,  regex:
        - CIM* matches any name that starts with CIM
        - CIM_abc* matches any name that starts with CIM_abc
        - CIM_ABC matches only the name CIM_ABC.
        - *ABC matches any name that includes ABC

    Parameters:

      pattern (:term:`string`):
        Python glob pattern to match.

      name_list:
        List of strings to be matched.

      ignore_case (bool):
        If True, do case-insensitive match. Default = True

    Returns:
        List of names that match.

    Raises:
        click.ClickException for regex compile error
    """

    flags = re.IGNORECASE if ignore_case else None
    # compile the regex since it used multiple times
    regex = None
    try:
        # Convert the glob input to regex.
        regex = fnmatch.translate(pattern)
        compiled_regex = re.compile(regex, flags)

    except Exception as ex:
        raise click.ClickException('Regex compile error. Regex={}. Er: {}: {}'
                                   .format(regex, ex.__class__.__name__, ex))

    new_list = [n for n in name_list for m in [compiled_regex.match(n)] if m]

    return new_list


def verify_operation(txt, msg=None):
    """
    Issue click confirm request and return result.  If msg is none and
    confirm response is n, output msg.

    Parameters:

      txt (:term:`string`):
        String that is prefixed to the prompt text and defines the
        verification request.

      msg (:class:`py:bool`):
        Optional parameter that if True causes an abort msg on the console.

    Returns:
        (:class:`py:bool`) where true corresponds to 'y' prompt response
    """
    if click.confirm(txt):
        return True
    if msg:
        click.echo('Request aborted')
    return False


def parse_wbemuri_str(wbemuri_str, namespace=None):
    """
    Parse a string that is a wbemuri into a CIMInstanceName object.  This method
    parses a string consistent with a wbemuri into a CIMInstanceName object.

    If the wbem_uri includes a namespace that is parsed also.  If both
    the namespace optional parameter exists and the wbemuri_str parameter
    includes a namespace (and they are not the same), an exception is returned.

    Returns:
        CIMInstanceName instance object

    Raises:
        ClickException: if the input wbemuri_str is an invalid wbemuri.
    """
    try:
        instance_name = CIMInstanceName.from_wbem_uri(wbemuri_str)
        if instance_name.namespace and namespace:
            if instance_name.namespace != namespace:
                raise click.ClickException('Conflicting namespaces between '
                                           'wbemuri {} and option {}'
                                           .format(instance_name.namespace,
                                                   namespace))
        elif instance_name.namespace is None and namespace:
            instance_name.namespace = namespace

        return instance_name
    except ValueError as ve:
        raise click.ClickException('Invalid wbem uri input {}. Error {}'
                                   .format(wbemuri_str, ve))


def create_cimvalue(cim_type, value_str, is_array):
    """
    Build a cim value of the type in cim_type and the information in value_str
    or fail with an exception if the value_str cannot be parsed into a
    CIMValue or list of CIMValue elements.

    Parameters:
      cim_type (:term:`string`):
        CIMType for this value. The CIM data type name for the CIM object.
        See :ref:`CIM data types` for valid type names.

      value_str (:term:`string`):
        String defining the input to be parsed.

      is_array (:class:`py:bool`):
        The value_str is to be treated as a comma separated set of values.

    Return:
        is_array == False. Returns a single CIMValue
        is_array == True. Returns a list of CIMValues

    Raises:
        ValueError if the value_str cannot be parsed consistent with
        the cim_type and is_array attributes of the call.
    """
    def str_2_bool(value):
        """
        Convert the value input to boolean based on text or
        raise ValueError if strings are not 'true' or 'false'.
        """
        if isinstance(value, bool):
            return value

        if isinstance(value, six.string_types):
            if value.lower() == 'true':
                return True
            if value.lower() == 'false':
                return False
        raise ValueError('Invalid boolean value: "{}"'.format(value))

    cim_value = None

    if not is_array:
        # cimvalue does not handle strings for bool
        if cim_type == 'boolean':
            value_str = str_2_bool(value_str)
        cim_value = cimvalue(value_str, cim_type)
    else:
        cim_value = []
        values_list = split_array_value(value_str, ',')
        for val in values_list:
            if cim_type == 'boolean':
                val = str_2_bool(val)
            cim_value.append(cimvalue(val, cim_type))
    return cim_value


def create_cimproperty(cim_type, is_array, name, value_str):
    """
    Create and return a CIMProperty from the input parameters and the
    information in cim_class.

    Parameters:

      cim_class (:class:`~pywbem.CIMClass`):
        CIM Class that includes the property defined by name

      name (:term:`string`):
        Name of the property to be constructed

      value_str (:term:`string`):
        String form for the value to be inserted.

    Returns:
        CIMProperty with name defined by name and CIMValue corresponding to
        value_str and property information from the class

    Raises:
        ValueError if value_str, cim_type and is_array mismatch.
    """
    cim_value = create_cimvalue(cim_type, value_str, is_array)

    return CIMProperty(name, cim_value, cim_type)


def create_ciminstancename(cim_class, kv_keys):
    """
    Create a CIMInstanceName object from the input parameters.

    The provided key values are verified against the properties of the class.

    Parameters:

      cim_class (:class:`~pywbem.CIMClass`):
        The class from which the CIMInstanceName is to be created.
        The provided keys are validated against the properties of that class.

      kv_keys (tuple):
        A tuple of name/value pairs representing the keys and their
        values that are to be constructed for the instance name. Required

    Returns:
        CIMInstanceName: with namespace = None and host = None

    Raises:
        click.ClickException if Property name not found in class or if mismatch
          of property type in class vs value component of kv pair
    """
    # TODO: Avoid the CIMClass object as input, because GetClass is not
    # implemented by all WBEM servers. That is not easy however, because the
    # name=value syntax used in the --key option does not allow distinguishing
    # the string/numeric/boolean types sufficiently well. For example, one
    # would need to find a way to resolve the ambiguity between numeric/boolean
    # values and strings with that same string value. One option might be to
    # have optional single or double quotes around string values, that normally
    # would not be needed, but when the string value is a numeric or boolean
    # string, then they would be used to turn the value into a string. For
    # example: name=42 -> numeric, name="42" -> string.
    keys = []
    for kv_property in kv_keys:
        name, value_str = parse_kv_pair(kv_property)
        try:
            cl_prop = cim_class.properties[name]
        except KeyError:
            raise click.ClickException('Property name "{}" not in class "{}".'
                                       .format(name, cim_class.classname))

        if value_str and value_str.startswith('"') and value_str.endswith('"'):
            value_str = value_str[1:-1]
        try:
            value = create_cimvalue(cl_prop.type, value_str, False)
            keys.append((name, value))
        except ValueError as ex:
            raise click.ClickException("Type mismatch property '{}' between "
                                       "expected type='{}', array={} and input "
                                       "value='{}'. Exception: {}"
                                       .format(name, cl_prop.type,
                                               cl_prop.is_array,
                                               value_str, ex))

    try:
        instance_path = CIMInstanceName(cim_class.classname, keybindings=keys)
    except ValueError as exc:
        raise click.ClickException(str(exc))

    return instance_path


def create_ciminstance(cim_class, kv_properties):
    """
    Create a CIMInstance object from the input parameters.

    Parameters:

      cim_class (CIMClass):
        The class from which the CIMInstance is to be created

      kv_properties (tuple):
        A tuple of name/value pairs representing the properties and their
        values that are to be constructed for the instance. Required

    Returns:
        CIMInstance

    Raises:
        click.ClickException if Property name not found in class or if mismatch
          of property type in class vs value component of kv pair
    """
    properties = []
    for kv_property in kv_properties:
        name, value_str = parse_kv_pair(kv_property)
        try:
            cl_prop = cim_class.properties[name]
        except KeyError:
            raise click.ClickException('Property name "{}" not in class "{}".'
                                       .format(name, cim_class.classname))

        try:
            prop = create_cimproperty(cl_prop.type,
                                      cl_prop.is_array,
                                      name,
                                      value_str)
            properties.append((name, prop))
        except ValueError as ex:
            raise click.ClickException("Type mismatch property '{}' between "
                                       "expected type='{}', array={} and input "
                                       "value='{}'. Exception: {}"
                                       .format(name, cl_prop.type,
                                               cl_prop.is_array,
                                               value_str, ex))

    new_inst = CIMInstance(cim_class.classname, properties=properties)

    return new_inst


def compare_obj(obj1, obj2, msg):
    """
    Compare two objects and display error if different.  Returns True if
    match or False if different
    """
    if obj1 != obj2:
        click.echo('Obj Compare {}: compare mismatch:\n{!r}\n{!r}'
                   .format(msg, obj1, obj2))
        return False
    return True


def compare_instances(inst1, inst2):
    """
    Compare two instances. If they do not match, compare the details to
    find differnes and report the differences. Report the differences
    """
    if inst1 != inst2:
        if not compare_obj(inst1.classname, inst2.classname, "classname"):
            return False
        if not compare_obj(inst1.path, inst2.path, "path"):
            return False
        if not compare_obj(inst1.qualifiers, inst2.qualifiers, "qualifiers"):
            return False
        if inst1.properties == inst2.properties:
            return True
        if len(inst1.properties) != len(inst2.properties):
            click.echo('Different number of properties {} vs {}\n{}\n{}'
                       .format(len(inst1.properties), len(inst2.properties),
                               inst1.keys(), inst2.keys()))
            return False
        keys1 = set(inst1.keys())
        keys2 = set(inst2.keys())
        if keys1 != keys2:
            diff = keys1.symmetric_difference(keys2)
            click.echo('Property Name differences {}'.format(diff))
            return False

        for n1, v1 in six.iteritems(inst1):
            if v1 != inst2[n1]:
                msg = 'property ' + n1
                if not compare_obj(inst1.get(n1), inst2.get(n1), msg):
                    return False
    return True


def parse_kv_pair(pair):
    """
    Parse a single key/value pair string in 'KEY=VALUE' syntax, and return a
    tuple (key, value).

    The parsing assumes that KEY does not include '=', which is always the case
    for CIM names. VALUE may contain '=' characters, which are retained.

    If 'VALUE' or '=VALUE' are missing, value is returned as None.
    """
    name, value = pair.partition("=")[::2]

    # if VALUE is missing, return it as None.
    if not value:
        value = None

    return name, value


def split_array_value(astring, delimiter):
    """Simple split of a string based on a delimiter"""

    rslt = split_str_w_esc(astring, delimiter)
    return rslt


def split_str_w_esc(astring, delimiter, escape='\\'):
    """
    Split string based on delimiter defined in call and the escape character \\
    To escape use of the delimiter in the strings. Delimiter may be multi
    character.
    Returns list of elements split from the input str
    """
    ret = []
    current_element = []
    iterator = iter(astring)
    for ch in iterator:
        if ch == escape:
            try:
                next_character = next(iterator)
                # Do not copy escape character if intended to escape either the
                # delimiter or the escape character itself. Copy the escape
                # character if it is not in use to escape one of these
                # characters.
                if next_character not in [delimiter, escape]:
                    current_element.append(escape)
                current_element.append(next_character)
            except StopIteration:
                current_element.append(escape)
        elif ch == delimiter:
            # split! (add current to the list and reset it)
            ret.append(''.join(current_element))
            current_element = []
        else:
            current_element.append(ch)
    ret.append(''.join(current_element))
    return ret


def get_cimtype(objects):
    """
    Get the cim_type for any returned cim object.  Normally this is the
    name of the class name except that the classname return from
    getclass and enumerate class is just unicode string
    """
    # associators and references return tuple
    if isinstance(objects, list):
        test_object = objects[0]
    elif objects:
        test_object = object
    else:
        cim_type = 'unknown'
        return None

    if isinstance(test_object, tuple):
        # associator or reference class level return is tuple
        cim_type = test_object[0].__class__.__name__
    else:
        cim_type = test_object.__class__.__name__

    # account for fact the enumerate class name operation returns uniicode.
    if isinstance(test_object, six.string_types):
        cim_type = 'CIMClassName'
    return cim_type


def process_invokemethod(context, objectname, methodname, options):
    # pylint: disable=line-too-long
    """
    Process the parameters for invokemethod at either the class or instance
    level and execute the invokemethod.

    Parameters:

      objectname (:term:`string` or :class:`~pywbem.CIMInstanceName` or :class:`~pywbem.CIMInstanceName`)  # noqa: E501

      methodname (:term:`string`):
        The name of the method to be executed

    """  # pylint: enable=line-too-long

    def create_params(classname, cim_method, kv_params):
        """
        Create parameter values from the input arguments and class.

        Parameters:

          cim_method (CIMMethod):
            CIM Method that is the template for the parameters.  It is used to
            evaluate the kv_params and generate corresponding CIMParameter
            objects to be passed to the InvokeMethod
        """
        params = []
        for p in kv_params:
            name, value_str = parse_kv_pair(p)
            if name not in cim_method.parameters:
                raise click.ClickException(
                    "Method {} of class {} does not have a parameter {}".
                    format(cim_method.name, classname, name))

            if name in params:
                raise click.ClickException(
                    "Method parameter {} specified multiple times".
                    format(name))

            cl_param = cim_method.parameters[name]
            is_array = cl_param.is_array

            cim_value = create_cimvalue(cl_param.type, value_str, is_array)
            params.append((name, cim_value))
        return params

    classname = objectname.classname \
        if isinstance(objectname, (CIMClassName, CIMInstanceName)) \
        else objectname

    cim_class = context.conn.GetClass(
        classname,
        namespace=options['namespace'], LocalOnly=False)

    cim_methods = cim_class.methods
    if methodname not in cim_methods:
        raise click.ClickException(
            "Class {} does not have a method {}"
            .format(classname, methodname))
    cim_method = cim_methods[methodname]

    params = create_params(classname, cim_method, options['parameter'])

    rtn = context.conn.InvokeMethod(methodname, objectname, params)

    # Output results, both ReturnValue and all output parameters
    click.echo('ReturnValue={}'.format(rtn[0]))

    if rtn[1]:
        cl_params = cim_method.parameters
        rtn_params = rtn[1]
        for pname, pvalue in rtn_params.items():
            ptype = cl_params[pname].type if pname in cl_params else None
            val = cimvalue_to_fmtd_string(
                pvalue, ptype, maxline=DEFAULT_MAX_CELL_WIDTH,
                avoid_splits=False)
            click.echo('{}={}'.format(pname, val[0]))


####################################################################
#
#  Display of CIM objects.
#
####################################################################


def sort_cimobjects(cim_objects):
    """
    Sort lists of CIMClass, CIMCLassName, CIMQualifierDecl, CIMInstance or
    CIMInstanceName. Sorts based on name or CIMInstancename. Sorting is based
    on the name value (name, classname, wbemuri(canonical form)
    Returns new list with the sorted objects.  This was defined as a common
    sort mechanism for all of the CIM object responses from WBEM servers.
    """
    if len(cim_objects) < 2:
        return cim_objects

    tst_obj = cim_objects[0]
    # this covers lists of classnames from class enum -o
    if isinstance(tst_obj, six.string_types):
        return sorted(cim_objects)

    if isinstance(tst_obj, CIMClass):
        return sorted(cim_objects, key=lambda class_: class_.classname)

    if isinstance(tst_obj, (CIMClassName, CIMInstanceName)):
        sort_dict = {obj.to_wbem_uri(format="canonical"): obj
                     for obj in cim_objects}
    elif isinstance(tst_obj, CIMInstance):
        sort_dict = {obj.path.to_wbem_uri(format="canonical"): obj
                     for obj in cim_objects}
    elif isinstance(tst_obj, CIMQualifierDeclaration):
        sort_dict = {obj.name: obj for obj in cim_objects}
    # Oddball case. In this case it is a tuple of CIMClassname,
    # CIMClass from class references/associators
    elif isinstance(tst_obj, tuple):
        assert isinstance(tst_obj[0], CIMClassName)
        assert isinstance(tst_obj[1], CIMClass)
        sort_dict = {tup[0].to_wbem_uri(format="canonical"): tup for tup in
                     cim_objects}
    else:
        raise TypeError('{} cannot be sorted'.format(type(cim_objects[0])))

    return [sort_dict[key] for key in sorted(sort_dict.keys())]


def display_cim_objects_summary(context, objects, output_format):
    """
    Display a summary of the objects received. This displays the
    count of objects.
    """
    context.spinner_stop()

    if objects:
        cim_type = get_cimtype(objects)

        if output_format_is_table(output_format):
            rows = [[len(objects), cim_type]]
            click.echo(format_table(rows, ['Count', 'CIM Type'],
                                    title='Summary of {} returned'
                                    .format(cim_type),
                                    table_format=output_format))
            return
        click.echo('{} {}(s) returned'.format(len(objects), cim_type))

    else:
        click.echo('0 objects returned')


def display_cim_objects(context, cim_objects, output_format, summary=False,
                        sort=False):
    """
    Display CIM objects in form determined by input parameters.

    Input is either a list of cim objects or a single object. It may be
    any of the CIM types.  This is used to display:

      * CIMClass

      * CIMClassName:

      * CIMInstance

      * CIMInstanceName

      * CIMQualifierDeclaration

      * Or list of the above

    This function may override output type choice in cases where the output
    choice is not available for the object type.  Thus, for example,
    mof output makes no sense for class names. In that case, the output is
    the str of the type.

    Parameters:

      context (:class:`ContextObj`):
        Click context contained in ContextObj object.

      objects (iterable of :class:`~pywbem.CIMInstance`,
        :class:`~pywbem.CIMInstanceName`, :class:`~pywbem.CIMClass`,
        :class:`~pywbem.CIMClassName`,
        or :class:`~pywbem.CIMQualifierDeclaration`):
        Iterable of zero or more CIM objects to be displayed.

      output_format (:term:`string`):
        String defining the preferred output format. Must not be None since
        the correct output_format must have been selected before this call.
        Note that the output formats allowed may depend on a) whether
        summary is True, b)the specific type because we do not have a table
        output format for CIMClass.

      summary (:class:`py:bool`):
        Boolean that defines whether the data in objects should be displayed
        or just a summary of the objects (ex. count of number of objects).
    """
    # Note: In the docstring above, the line for parameter 'objects' was way too
    #       long. Since we are not putting it into docmentation, we folded it.

    context.spinner_stop()

    if summary:
        display_cim_objects_summary(context, cim_objects, output_format)
        return

    if not cim_objects and context.verbose:
        click.echo("No objects returned")
        return

    if sort:
        cim_objects = sort_cimobjects(cim_objects)

    # default when displaying cim objects is mof
    assert output_format

    if isinstance(cim_objects, (list, tuple)):
        # Table format output is processed as a group
        if output_format_is_table(output_format):
            _print_objects_as_table(cim_objects, output_format, context=context)
        else:
            # Call to display each object
            for obj in cim_objects:
                display_cim_objects(context, obj, output_format=output_format)
        return

    # Display a single item.
    object_ = cim_objects
    # This allows passing single objects to the table formatter (i.e. not lists)
    if output_format_is_table(output_format):
        _print_objects_as_table([object_], output_format, context=context)
    elif output_format == 'mof':
        try:
            click.echo(object_.tomof())
        except AttributeError:
            # insert NL between instance names for readability
            if isinstance(object_, CIMInstanceName):
                click.echo("")
                click.echo(object_)
            elif isinstance(object_, (CIMClassName, six.string_types)):
                click.echo(object_)
            else:
                raise click.ClickException('output_format {} invalid for {} '
                                           .format(output_format,
                                                   type(object_)))
    elif output_format == 'xml':
        try:
            click.echo(object_.tocimxmlstr(indent=4))
        except AttributeError:
            # no tocimxmlstr functionality
            raise click.ClickException('Output Format {} not supported. '
                                       'Default to\n{!r}'
                                       .format(output_format, object_))
    elif output_format == 'repr':
        try:
            click.echo(repr(object_))
        except AttributeError:
            raise click.ClickException('"repr" display of {!r} failed'
                                       .format(object_))

    elif output_format == 'txt':
        try:
            click.echo(object_)
        except AttributeError:
            raise click.ClickException('"txt" display of {!r} failed'
                                       .format(object_))
    # elif output_format == 'tree':
    #    raise click.ClickException('Tree output format not allowed')
    else:
        raise click.ClickException('Invalid output format {}'
                                   .format(output_format))


def _print_classes_as_table(classes, table_width, table_format):
    """
    TODO: Future extend to display classes as a table, showing the
    properties for each class. This will display the properties that exist in
    subclasses. The temp output
    so we could create the function is to just output as mof
    """
    # pylint: disable=unused-argument

    for class_ in classes:
        click.echo(class_.tomof())


def format_keys(obj, max_width):
    """
    Format the keys of a dictionary of keybindings as text for display. Formats
    multiple keybindings on each line within the max_width

    Parameters:

      obj (:class:`pwbem.CIMInstanceName`):
        Instance name from which keybindings are to be extracted for
        formatting.

    Returns:
        :term:`string` containing the keys from the input obj formatted for
        display at within the defined width.
    """
    def get_wbemurikeys(obj):
        """
        Create wbem_uri from CIMInstanceName and separate out key component
        for return.
        """
        wbem_uri = obj.to_wbem_uri()
        wbem_uri_keys = wbem_uri[wbem_uri.find('.'):]
        wbem_uri_keys = wbem_uri_keys[1:]
        return wbem_uri_keys

    assert isinstance(obj, CIMInstanceName)
    # clear the host and namespace
    myobj = obj.copy()
    myobj.host = None
    myobj.namespace = None
    wbem_uri_keys = get_wbemurikeys(myobj)

    # Too long for width. Fold the keys on multiple lines
    if len(wbem_uri_keys) > max_width:
        wbem_uri_keys = ''
        line_len = 0
        for key, value in myobj.keybindings.items():
            one_key_obj = get_wbemurikeys((CIMInstanceName('x', {key: value})))
            if wbem_uri_keys:
                if line_len + len(one_key_obj) > max_width:
                    wbem_uri_keys += '\n{}'.format(one_key_obj)
                    line_len = 0
                else:
                    wbem_uri_keys += ',{}'.format(one_key_obj)
                    line_len += len(one_key_obj) + 1

            else:  # must put on first line even if too long
                wbem_uri_keys += one_key_obj
                line_len = len(one_key_obj) + 1

    return wbem_uri_keys


def display_text(text, output_format=None):  # pylint: disable=unused-argument
    """
    Display the text output format. Currently this simply outputs to
    click.echo
    """
    click.echo(text)


class NoCaseList(object):
    """
    This class simplifies working with lists of strings that are to be tested
    case-insenstive. This simplifies working with  the CIM object attributes
    that are case insensitive such as class names, roles, etc.

    NOTE: This code does not handle None either for items in the list or for
    the str parameter in __contains__

    Example:

        roles = NoCaseList(roles)
        # the following does case insensitive test
        if role in roles:
            do something
    """

    def __init__(self, strs):
        """
        Constructor requires list of strings input

        Parameters:

          strs (list of :term:`string`):
            The strings that will make up the list
        """

        assert isinstance(strs, list)
        self.str_list = strs

    def __contains__(self, astring):
        """
        Implement Python 'in' functionality.

        Parameters:

          astring (:term:`string`):
            String to test against the instance of NocaseList

        Returns:
          True if string in list (case insensitive compare)
        """

        assert astring is not None
        return astring.lower() in (n.lower() for n in self.str_list)

    def add(self, strs):
        """
        Add string or list of strings to the list

        Parameters:

          strs (:term:`string` or list, tuple of :term:`string`):
            String to test against the instance of NocaseList.
      """
        if not strs:
            return
        if isinstance(strs, list):
            self.str_list.extend(strs)
        else:
            self.str_list.append(strs)

    def get(self, astring):
        """
        Get the string that matches astring case insensitive

        Parameters:

          astring (:term:`string` or list, tuple of :term:`string`):
            String to test against the instance of NocaseList.

        Returns:
            String in list that matches No case the input string, or None
            if the input string was empty or None.

        Raises:
            KeyError: If no match
        """
        if not astring:
            return None
        tstr = astring.lower()
        for s in self.str_list:
            if tstr == s.lower():
                return s
        raise KeyError('{} not found in list'.format(astring))


def shorten_path_str(path, replacements, fullpath):
    """
    Create a short-form path str from the input CIMInstanceName with selected
    components shortened to just a single known character.  This allows
    modifying the path string to replace selected key/value paris with a single
    character. Thus where the original string is very long and contains
    repeated key bindings (ex. CreationClassName) we can shorten the path
    string by reducing selected key/value pairs to just ~

    Parameters:

      path (:class:`CIMInstanceName`):
        CIMInstanceName object defining instance name to shorten

      replacements:
        Dictionary of the replacements containing a key names and key
        values to be replaced. If the key value is None, they name alone
        causes the replacement. Otherwise, both the name and value must
        match.

      fullpath (:class:`py:bool`):
        If True Return complete path using to_wbem_rul. Otherwise, shorten
        the path by replacing keys defined by the replacements dictionary.
        shorten the path, otherwise simply convert to string. Othewise

    Returns:
        String representation of the path.
    """

    if fullpath:
        # Just build the full path string
        name_str = path.to_wbem_uri()

    else:
        # Shorten path based on key definitons in replacements
        kbs = path.keybindings
        repl_list = []
        magicvalue = 9999123999918
        for k, v in kbs.items():
            for key, value in replacements.items():
                if k.lower() == key.lower():
                    if value is None or v == value:
                        repl_list.append((key, value))
                        # Set the value to a known value for the replacement
                        kbs[key] = magicvalue
        path.keybindings = kbs
        name_str = path.to_wbem_uri()
        # replace each key binding in repl_list with ~ char
        for key, value in repl_list:
            name_str = name_str.replace("{}={}".format(key, magicvalue), "~", 1)

    return name_str


def _print_paths_as_table(objects, table_width, table_format):
    # pylint: disable=unused-argument
    """
    Display paths as a table. This include CIMInstanceName, ClassPath,
    and unicode (the return type for enumerateClasses).
    """
    title = None
    if objects:
        if isinstance(objects[0], six.string_types):
            title = 'Classnames:'
            headers = ['Class Name']
            rows = [[obj] for obj in objects]
        elif isinstance(objects[0], CIMClassName):
            title = 'Classnames'
            headers = ('host', 'namespace', 'class')
            rows = [[obj.host, obj.namespace, obj.classname] for obj in objects]
        elif isinstance(objects[0], CIMInstanceName):
            title = 'InstanceNames: {}'.format(objects[0].classname)
            host_hdr = 'host'
            ns_hdr = 'namespace'
            class_hdr = 'class'
            host_hdr_len = len(host_hdr) + 4
            ns_hdr_len = len(ns_hdr) + 3
            class_hdr_len = len(class_hdr) + 3
            headers = (host_hdr, ns_hdr, class_hdr, 'keysbindings')

            host_lens = [len(obj.host) for obj in objects if obj.host]
            host_max = max(host_lens) if host_lens else host_hdr_len
            ns_lens = [len(obj.namespace) for obj in objects if obj.namespace]
            ns_max = max(ns_lens) if ns_lens else ns_hdr_len
            class_lens = [len(obj.classname) for obj in objects]
            class_max = max(class_lens) if class_lens else class_hdr_len

            max_key_len = (table_width) - (host_max + ns_max + class_max + 3)
            rows = [[obj.host, obj.namespace, obj.classname,
                     format_keys(obj, max_key_len)] for obj in objects]
        else:
            raise click.ClickException("{0} invalid type ({1})for path display".
                                       format(objects[0], type(objects[0])))

        click.echo(format_table(rows, headers, title=title,
                                table_format=table_format))


def _print_qual_decls_as_table(qual_decls, table_width, table_format):
    """
    Display the elements of qualifier declarations as a table with a
    row for each qualifier declaration and a column for each of the attributes
    of the qualifier declaration (name, type, Value, Array, Scopes, Flavors.

    The function displays all of the qualifier declarations in the
    """
    rows = []
    headers = ['Name', 'Type', 'Value', 'Array', 'Scopes', 'Flavors']
    max_column_width = (table_width / len(headers)) - 4
    for q in qual_decls:
        scopes = '\n'.join([key for key in q.scopes if q.scopes[key]])
        flavors = []
        flavors.append('EnableOverride' if q.overridable else 'DisableOverride')
        flavors.append('ToSubclass' if q.tosubclass else 'Restricted')
        if q.translatable:
            flavors.append('Translatable')
        if sum([len(i) for i in flavors]) > max_column_width:
            sep = "\n"
        else:
            sep = ", "
        flavors = sep.join(flavors)

        row = [q.name, q.type, q.value, q.is_array, scopes, flavors]
        rows.append(row)

    click.echo(format_table(rows, headers, title='Qualifier Declarations',
                            table_format=table_format))


def _format_instances_as_rows(insts, max_cell_width=DEFAULT_MAX_CELL_WIDTH,
                              include_classes=False, context=None,
                              prop_names=None):
    """
    Format the list of instances properties into as a list of the property
    values for each instance( a row of the table) gathered into a list of
    the rows.

    The prop_names parameter is the list of (originally cased) property names
    to be output, in the desired output order. It could be determined from
    the instances, but since it is determined already by the caller, it
    is passed in as an optimization. For test convenience, None is permitted
    and causes the properties to again be determined from the instances.

    Include_classes for each instance if True. Sets the classname as the first
    column.

    max_width if not None folds col entries longer than the defined
    max_cell_width. If max_width is None, the data length is ignored.

    The property values are formatted similar to MOF output. Properties that
    have a ValueMap qualifier (effectively, in the creation class of the
    instance) are shown with both the actual property value and the mapped
    value in parenthesis.

    NOTE: This is a separate function to allow testing of the table formatting
    independently of print output.

    Returns:
        list of strings where each string is a row in the table and each
        item in a row is a cell entry
    """
    # Avoid crash deeper in code if max_cell_width is None.
    if max_cell_width is None:
        max_cell_width = DEFAULT_MAX_CELL_WIDTH
    lines = []

    if prop_names is None:
        prop_names = sorted_prop_names(insts)

    # Cache of ValueMapping objects for integer-typed properties.
    # Key: classname.propertyname, both in lower case.
    # A value of None indicates the property does not have a value mapping.
    valuemappings = {}

    for inst in insts:
        if not isinstance(inst, CIMInstance):
            raise ValueError('Only accepts CIMInstance; not type {}'
                             .format(type(inst)))

        # Insert classname as first col if flag set
        line = [inst.classname] if include_classes else []

        # get value for each property in this object
        for name in prop_names:

            # Account for possible instances without all properties
            # Outputs empty  string.  Note that instance with no value
            # results in same output as not instance name.
            if name not in inst.properties:
                val_str = ''
            else:
                value = inst.get(name)
                p = inst.properties[name]

                # Cache value mappings for integer-typed properties
                if INT_TYPE_PATTERN.match(p.type) and context:
                    vm_key = '{}.{}'.format(
                        inst.classname.lower(), name.lower())
                    try:
                        valuemapping = valuemappings[vm_key]
                    except KeyError:
                        try:
                            valuemapping = ValueMapping.for_property(
                                context.conn,
                                context.conn.default_namespace,
                                inst.classname,
                                name)
                        except ValueError:
                            # Property does not have a value mapping.
                            valuemapping = None
                        valuemappings[vm_key] = valuemapping
                else:
                    valuemapping = None

                if value is None:
                    val_str = u''
                else:
                    val_str, _ = cimvalue_to_fmtd_string(
                        p.value, p.type, indent=0, maxline=max_cell_width,
                        line_pos=0, end_space=0, avoid_splits=False,
                        valuemapping=valuemapping)

            line.append(val_str)
        lines.append(line)

    return lines


def _print_instances_as_table(insts, table_width, table_format,
                              include_classes=False, context=None):
    """
    Print the properties of the instances defined in insts as a table where
    each row is an instance and each column is a property value.

    All properties in the instance are included.

    The header line consists of the property names.

    The property values are formatted similar to MOF output. Properties that
    have a ValueMap qualifier (effectively, in the creation class of the
    instance) are shown with both the actual property value and the mapped
    value in parenthesis.
    """

    if table_width is None:
        table_width = DEFAULT_TABLE_WIDTH

    prop_names = sorted_prop_names(insts)

    # Try to estimate max cell width from number of cols
    # This allows folding long data.  However it is incomplete in
    # that we do not fold the property name.  Further, the actual output
    # width of a column involves the tabulate outputter, output_format
    # so this is not deterministic.
    if prop_names:
        num_cols = len(prop_names)
        max_cell_width = int(table_width / num_cols) - 2
    else:
        max_cell_width = table_width

    header_line = []
    if include_classes:
        header_line.append("classname")
    header_line.extend(prop_names)

    # Fold long property names
    new_header_line = []
    for header in header_line:
        if len(header) > max_cell_width:
            new_header_line.append(fold_strings(header, max_cell_width))
        else:
            new_header_line.append(header)

    for inst in insts:
        if not isinstance(inst, CIMInstance):
            raise ValueError('Only CIMInstance display allows table output')

    rows = _format_instances_as_rows(insts, max_cell_width=max_cell_width,
                                     include_classes=include_classes,
                                     context=context, prop_names=prop_names)

    title = 'Instances: {}'.format(insts[0].classname)
    click.echo(format_table(rows, new_header_line, title=title,
                            table_format=table_format))


def sorted_prop_names(insts):
    """
    Return the list of (originally cased) property names that is the superset
    of all properties in the input instances.

    The returned list has the key properties first, followed by the non-key
    properties. Each group is sorted case insensitively.

    The key properties are determined from the instance paths, if present.
    The function tolerates it if only some of the instances have a path,
    and if instances of subclasses have additional keys.
    """

    all_props = odicti()  # key: org prop name, value: lower cased prop name
    key_props = odicti()  # key: org prop name, value: lower cased prop name
    for inst in insts:
        inst_props = inst.keys()
        for pn in inst_props:
            all_props[pn] = pn.lower()
        if inst.path:
            key_prop_names = inst.path.keys()
            for pn in inst_props:
                if pn in key_prop_names:
                    key_props[pn] = pn.lower()

    nonkey_props = odicti()  # key: org prop name, value: lower cased prop name
    for pn in all_props:
        if pn not in key_props:
            nonkey_props[pn] = all_props[pn]

    key_prop_list = sorted(key_props.keys(), key=lambda p: p.lower())
    nonkey_prop_list = sorted(nonkey_props.keys(), key=lambda p: p.lower())
    key_prop_list.extend(nonkey_prop_list)
    return key_prop_list


def _print_objects_as_table(objects, output_format, context=None):
    """
    Call the method for each type of object to print that object type
    information as a table.

    Output format is retrieved from context.
    """
    if USE_TERMINAL_WIDTH:
        table_width = click.get_terminal_size()[0]
    else:
        table_width = DEFAULT_TABLE_WIDTH

    if objects:
        if isinstance(objects[0], CIMInstance):
            _print_instances_as_table(objects, table_width, output_format,
                                      context=context)
        elif isinstance(objects[0], CIMClass):
            _print_classes_as_table(objects, table_width, output_format)
        elif isinstance(objects[0], CIMQualifierDeclaration):
            _print_qual_decls_as_table(objects, table_width, output_format)
        elif isinstance(objects[0], (CIMClassName, CIMInstanceName,
                                     six.string_types)):
            _print_paths_as_table(objects, table_width, output_format)
        else:
            raise click.ClickException("Cannot print {} as table"
                                       .format(type(objects[0])))


def hide_empty_columns(headers, rows):
    """
    Removes columns from rows if the colmuns are considered empty.
    The definiton of an empty row is:
    1. All entries for the column in all rows are None or "" if type string.
    2. All entries for the column in all rows are None if number.

    Returns new rows and headers
    """
    def column_is_empty(rows, column):
        """
        Determine if entries for defined column in all rows are considered
        empty.
        Returns True if all are empty. Otherwise returns False
        """
        for row in rows:
            if isinstance(row[column], six.integer_types) and \
                    row[column] is not None:
                return False
            if row[column]:
                return False
        return True

    # Remove empty rows
    len_hdr = len(headers)
    for row in rows:
        assert len(row) == len_hdr, "row: {}\nhdrs: {}". \
            format(row, headers)
    for column in range(len(headers) - 1, -1, -1):
        if column_is_empty(rows, column):
            del headers[column]
            for row in rows:
                del row[column]

    return headers, rows


def format_table(rows, headers, title=None, table_format='simple',
                 sort_columns=None):
    """
    General print table function.  Prints a list of lists in a
    table format where each inner list is a row.
    This code is temporary while the tabulate package is updated

    Parameters:

      headers (list strings):
        where each string is a table column name or None if no header is to be
        attached

      table_data (list of lists):
        where each the top level iterables represents the list of rows
        and each row is an iterable of strings for the data in that row.

      title (:term:`string`):
         Optional title to be places io the output above the table.
         No title is output if this parameter is None

      table_format (:term:`string`):
        Output format defined by the string and limited to one of the
        choice of table formats defined in TABLE_FORMATS list

      output_file (:term:`string`):
        If not None, a file name to which the output formatted data is sent.

      sort_columns (int or list of int that defines sort):
        Defines the cols that will be sorted. If int, it defines the column
        that will be sorted. If list of int, the sort is in sort order of
        cols in the list (i.e. minor sorts to the left, major sorts to the
        right). Note that entries in each row of the columns to be sorted
        must be of the same type (int, str, etc.) to be sortable.

    Returns:
        :term:`string`: Returns the formatted table as a string

    Raises:
        click.ClickException if invalid table format string
    """
    if sort_columns is not None:
        if isinstance(sort_columns, int):
            rows = sorted(rows, key=itemgetter(sort_columns))
        elif isinstance(sort_columns, (list, tuple)):
            rows = sorted(rows, key=itemgetter(*sort_columns))
        else:
            assert False, "Sort_columns must be int or list/tuple of int"

    if table_format is None:
        table_format = 'table'
    if table_format == 'table':
        table_format = 'psql'
    if not output_format_is_table(table_format):
        raise click.ClickException('Invalid table format {}.'
                                   .format(table_format))

    result = tabulate.tabulate(rows, headers, tablefmt=table_format)
    if title:
        if table_format == 'html':
            result = '<p>{0}</p>\n{1}'.format(title, result)
        else:
            result = '{0}\n{1}'.format(title, result)
    return result


def fold_strings(input_strings, max_width, break_long_words=False,
                 break_on_hyphens=False, fold_list_items=False, separator=', ',
                 initial_indent='', subsequent_indent=''):
    """
    Fold a string or a list/tuple of strings within a maximum width and return
    a folded string that fits within the width defined by max_width. If input
    is a list of strings, the fold_list_itmes defines whether they create
    separate lines in the output or are concatenated into a single string
    before folding the string. This implementation refolds strings that already
    contain EOL characters and removes any existing folds that do not match the
    max_width criteria.  Lists of strings may be folded with one string per
    line or concatenated and then folded.

    Parameters:

      input_strings (:term:`string` or list of :term:`string`):
        The string that will be contents of into the cell. This string may
        already include multiple lines.

      max_width (:term:`integer`):
        Maximum width of cell containing the resulting string.  Data is
        folded into multiple lines to fit into this width.

      break_long_words (:class:`py:bool`):
        Boolean that forces long words to be broken if True
        If False, long words will not break at the max width

      break_on_hyphens (:class:`py:bool`):
        If True use hypens as word separator.

      fold_list_items (:class:`py:bool`):
        If True, force fold for each item in list/tupe of input strings if
        single line concatenation is longer than max_width. Otherwise list is
        contatenated and the result folded.

      separator (:term:`string`):
        String that is separator between list items when folded into
        string.

      initial_indent (:term:`integer` or :term:`string`):
        Integer  or string defining the number of characters of indent for
        the first line if single string or the first line of each string in
        the list if input_strings is a list and fold_list_items is True.

      subsequent_indent (:term:`integer`):
        Integer or string, defining the number of characters of indent for
        the all but the first lineif single string or the first line of
        each string in the list if input_strings is a list and
        fold_list_items is True.

    Returns:
        String representing the folded input_strings
    """
    def indent_str(indent):
        """
        Maps indent to string of indent characters if necessary.
        If None, return None
        If integer, map to indent characters
        """
        if indent is None:
            return ""
        if isinstance(indent, six.string_types):
            return indent
        return ' ' * indent

    initial_indent = indent_str(initial_indent)
    subsequent_indent = indent_str(subsequent_indent)

    if isinstance(input_strings, (list, tuple)):
        if separator is None:
            separator = ', '
        build_str = separator.join(input_strings)
        # Return contatentated list if within cell width
        if len(build_str) <= max_width:
            return build_str

        if fold_list_items:
            folded_strings = []
            for str_item in input_strings:
                folded_strings.append(
                    fold_strings(str_item, max_width,
                                 break_long_words=break_long_words,
                                 break_on_hyphens=break_on_hyphens,
                                 initial_indent=initial_indent,
                                 subsequent_indent=subsequent_indent))
            return "\n".join(folded_strings)
        input_strings = build_str

    # process single string
    input_string = input_strings
    assert isinstance(input_string, six.string_types)

    if len(input_string) <= max_width:
        return input_string

    # use textwrap fill to fold the string
    folded_string = fill(input_string, max_width,
                         break_long_words=break_long_words,
                         break_on_hyphens=break_on_hyphens,
                         initial_indent=initial_indent,
                         subsequent_indent=subsequent_indent)

    return folded_string


def raise_pywbem_error_exception(er):
    """
    Raise the standard click exception for a pywbem Error exception.  These
    exceptions do not cause interactive mode failure but display the exception
    class and its str value and return to the repl mode.
    """
    raise click.ClickException('{}: {}'.format(er.__class__.__name__, er))



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Click command definition for the pywbemcli command, the top level command for
the pywbemcli click tool
"""

from __future__ import absolute_import, print_function

import os
import sys
import traceback
from copy import deepcopy
import click
import click_repl
from prompt_toolkit.history import FileHistory
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory

from pywbem import LOGGER_SIMPLE_NAMES, \
    LOG_DESTINATIONS, DEFAULT_LOG_DESTINATION, LOG_DETAIL_LEVELS, \
    DEFAULT_LOG_DETAIL_LEVEL
from pywbem import __version__ as pywbem_version

from ._context_obj import ContextObj, display_click_context
from ._common import GENERAL_OPTS_TXT, SUBCMD_HELP_TXT, OUTPUT_FORMAT_GROUPS
from ._common_options import add_options, help_option
from ._pywbem_server import PywbemServer
from .config import DEFAULT_NAMESPACE, \
    PYWBEMCLI_PROMPT, PYWBEMCLI_HISTORY_FILE, DEFAULT_MAXPULLCNT, \
    DEFAULT_CONNECTION_TIMEOUT, MAX_TIMEOUT, USE_AUTOSUGGEST
from ._connection_repository import ConnectionRepository
from ._click_extensions import PywbemcliTopGroup
from ._utils import deprecation_warning

__all__ = ['cli']


# Defaults for some options
DEFAULT_VERIFY = True  # The default is to verify
DEFAULT_TIMESTATS = False
DEFAULT_PULL_CHOICE = 'either'
USE_PULL_CHOICE = {'either': None, 'yes': True, 'no': False}

DEFAULT_CONNECTIONS_FILE = 'pywbemcli_connection_definitions.yaml'
DEFAULT_CONNECTIONS_PATH = os.path.join(os.path.expanduser("~"),
                                        DEFAULT_CONNECTIONS_FILE)

TERMWIDTH_ENVVAR = os.getenv(PywbemServer.termwidth_envvar, None)
if TERMWIDTH_ENVVAR:
    try:
        TERMWIDTH_ENVVAR = int(TERMWIDTH_ENVVAR)
    except ValueError:
        TERMWIDTH_ENVVAR = None

CONTEXT_SETTINGS = dict(

    # Enable -h as additional help option:
    help_option_names=['-h', '--help'],

    # Default the output width properly:
    terminal_width=TERMWIDTH_ENVVAR or click.get_terminal_size()[0],
)


def validate_connections_file(connections_repo):
    """
    Test for existence of a connections file and abort if it does not exist.
    Abort click if file does not exist
    """
    if not connections_repo.file_exists():
        click.echo('Connections file: "{}" does not exist.'.format(
            connections_repo.connections_file), err=True)
        raise click.Abort()


# pylint: disable=bad-continuation
# PywbemcliTopGroup sets order commands listed in help output
@click.group(invoke_without_command=True, cls=PywbemcliTopGroup,
             context_settings=CONTEXT_SETTINGS,
             options_metavar=GENERAL_OPTS_TXT,
             subcommand_metavar=SUBCMD_HELP_TXT)
@click.option('-n', '--name', 'connection_name', type=str, metavar='NAME',
              # defaulted in code
              envvar=PywbemServer.name_envvar,
              help=u'Use the WBEM server defined by the WBEM connection '
                   u'definition NAME. '
                   u'This option is mutually exclusive with the --server and '
                   u'--mock-server options, since each defines a WBEM server. '
                   u'Default: EnvVar {ev}, or none.'.
                   format(ev=PywbemServer.name_envvar))
@click.option('-m', '--mock-server', type=str, multiple=True, metavar="FILE",
              # defaulted in code
              envvar=PywbemServer.mock_server_envvar,
              help=u'Use a mock WBEM server that is automatically created in '
                   u'pywbemcli and populated with CIM objects that are defined '
                   u'in the specified MOF file or Python script file. '
                   u'See the pywbemcli documentation for more information. '
                   u'This option may be specified multiple times, and is '
                   u'mutually exclusive with the --server and --name options, '
                   'since each defines a WBEM server. '
                   'Default: EnvVar {ev}, or none.'.
                   format(ev=PywbemServer.mock_server_envvar))
@click.option('-s', '--server', type=str, metavar='URL',
              # defaulted in code
              envvar=PywbemServer.server_envvar,
              help=u'Use the WBEM server at the specified URL with format: '
                   u'[SCHEME://]HOST[:PORT]. '
                   u'SCHEME must be "https" (default) or "http". '
                   u'HOST is a short or long hostname or literal IPV4/v6 '
                   u'address. '
                   u'PORT defaults to 5989 for https and 5988 for http. '
                   u'This option is mutually exclusive with the --mock-server '
                   u'and --name options, since each defines a WBEM server. '
                   u'Default: EnvVar {ev}, or none.'.
                   format(ev=PywbemServer.server_envvar))
@click.option('-u', '--user', type=str, metavar='TEXT',
              # defaulted in code
              envvar=PywbemServer.user_envvar,
              help=u'User name for the WBEM server. '
                   u'Default: EnvVar {ev}, or none.'.
                   format(ev=PywbemServer.user_envvar))
@click.option('-p', '--password', type=str, metavar='TEXT',
              # defaulted in code
              envvar=PywbemServer.password_envvar,
              help=u'Password for the WBEM server. '
                   u'Default: EnvVar {ev}, or prompted for if --user '
                   u'specified.'.
                   format(ev=PywbemServer.password_envvar))
@click.option('--verify/--no-verify', 'verify', default=None,
              # defaulted in code
              envvar=PywbemServer.verify_envvar,
              help=u'If --verify, client verifies the X.509 server '
                   u'certificate presented by the WBEM server during TLS/SSL '
                   u'handshake. If --no-verify client bypasses verification. '
                   u'Default: EnvVar {ev}, or "--verify".'.
                   format(ev=PywbemServer.verify_envvar))
@click.option('--ca-certs', type=str, metavar="CACERTS",
              default=None,  # defaulted in code
              envvar=PywbemServer.ca_certs_envvar,
              help=u'Certificates used to validate the certificate presented '
                   u'by the WBEM server during TLS/SSL handshake: '
                   u'FILE: Use the certs in the specified PEM file; '
                   u'DIR: Use the certs in the PEM files in the specified '
                   u'directory; '
                   u'"certifi" (pywbem 1.0 or later): Use the certs provided '
                   u'by the certifi Python package; '
                   u'Default: EnvVar {ev}, or "certifi" (pywbem 1.0 or later), '
                   u'or the certs in the PEM files in the first existing '
                   u'directory from from a system defined list of directories '
                   u'(pywbem before 1.0).'.
                   format(ev=PywbemServer.ca_certs_envvar))
@click.option('-c', '--certfile', type=str, metavar="FILE",
              # defaulted in code
              envvar=PywbemServer.certfile_envvar,
              help=u'Path name of a PEM file containing a X.509 client '
                   u'certificate that is used to enable TLS/SSL 2-way '
                   u'authentication by presenting the certificate to the '
                   u'WBEM server during TLS/SSL handshake. '
                   u'Default: EnvVar {ev}, or none.'.
                   format(ev=PywbemServer.certfile_envvar))
@click.option('-k', '--keyfile', type=str, metavar='FILE',
              # defaulted in code
              envvar=PywbemServer.keyfile_envvar,
              help=u'Path name of a PEM file containing a X.509 private key '
                   u'that belongs to the certificate in the --certfile file. '
                   u'Not required if the private key is part of the '
                   u'--certfile file. '
                   u'Default: EnvVar {ev}, or none.'.
                   format(ev=PywbemServer.keyfile_envvar))
@click.option('-t', '--timeout', type=click.IntRange(0, MAX_TIMEOUT),
              metavar='INT',
              # defaulted in code
              envvar=PywbemServer.timeout_envvar,
              help=u'Client-side timeout in seconds for operations with the '
                   u'WBEM server. '
                   u'Default: EnvVar {ev}, or {default}.'.
                   format(ev=PywbemServer.timeout_envvar,
                          default=DEFAULT_CONNECTION_TIMEOUT))
@click.option('-U', '--use-pull', type=click.Choice(['yes', 'no', 'either']),
              # defaulted in code
              envvar=PywbemServer.use_pull_envvar,
              help=u'Determines whether pull operations are used for '
                   u'operations with the WBEM server that return lists of '
                   u'instances, as follows: '
                   u'"yes" uses pull operations and fails if not supported by '
                   u'the server; '
                   u'"no" uses traditional operations; '
                   u'"either" (default) uses pull operations if supported by '
                   u'the server, and otherwise traditional operations. '
                   u'Default: EnvVar {ev}, or "either".'.
                   format(ev=PywbemServer.use_pull_envvar))
@click.option('--pull-max-cnt', type=int, metavar='INT',
              # defaulted in code
              envvar=PywbemServer.pull_max_cnt_envvar,
              help=u'Maximum number of instances to be returned by the WBEM '
                   u'server in each open or pull response, if pull operations '
                   u'are used. '
                   u'This is a tuning parameter that does not affect the '
                   u'external behavior of the commands. '
                   u'Default: EnvVar {ev}, or {default}'.
                   format(ev=PywbemServer.pull_max_cnt_envvar,
                          default=DEFAULT_MAXPULLCNT))
@click.option('-T', '--timestats', is_flag=True,
              # defaulted in code
              help=u'Show time statistics of WBEM server operations.')
@click.option('-d', '--default-namespace', type=str, metavar='NAMESPACE',
              default=None,
              envvar=PywbemServer.defaultnamespace_envvar,
              help=u'Default namespace, to be used when commands do not '
                   u'specify the --namespace command option. '
                   u'Default: EnvVar {ev}, or {default}.'.
                   format(ev=PywbemServer.defaultnamespace_envvar,
                          default=DEFAULT_NAMESPACE))
@click.option('-o', '--output-format', metavar='FORMAT',
              help=u'Output format for the command result. '
                   u'The default and allowed output formats are command '
                   u'specific. '
                   u'The default output_format is None so that each command '
                   u'selects its own default format. '
                   u'FORMAT is: table formats: [{tb}]; CIM object '
                   u'formats: [{ob}]]; TEXT formats: [{tx}].'.
                   format(tb='|'.join(OUTPUT_FORMAT_GROUPS['TABLE'][0]),
                          ob='|'.join(OUTPUT_FORMAT_GROUPS['CIM'][0]),
                          tx='|'.join(OUTPUT_FORMAT_GROUPS['TEXT'][0])))
@click.option('-l', '--log', type=str, metavar='COMP[=DEST[:DETAIL]],...',
              # defaulted in code
              envvar=PywbemServer.log_envvar,
              help=u'Enable logging of the WBEM operations, defined by a list '
                   u'of log configuration strings with: '
                   u'COMP: [{comp_choices}]; '
                   u'DEST: [{dest_choices}], default: {dest_default}; '
                   u'DETAIL: [{detail_choices}], default: {detail_default}. '
                   u'Default: EnvVar {ev}, or {default}.'.
                   format(comp_choices='|'.join(LOGGER_SIMPLE_NAMES),
                          dest_choices='|'.join(LOG_DESTINATIONS),
                          dest_default=DEFAULT_LOG_DESTINATION,
                          detail_choices='|'.join(LOG_DETAIL_LEVELS),
                          detail_default=DEFAULT_LOG_DETAIL_LEVEL,
                          ev=PywbemServer.log_envvar,
                          default='all'))
@click.option('-v', '--verbose/--no-verbose',
              default=None,
              help=u'Display extra information about the processing.')
@click.option('--deprecation-warnings/--no-deprecation-warnings', is_flag=True,
              default=True,
              envvar=PywbemServer.deprecation_warnings_envvar,
              help=u'Enable deprecation warnings. '
              'Default: EnvVar {ev}, or true.'.
              format(ev=PywbemServer.deprecation_warnings_envvar))
@click.option('-C', '--connections-file', metavar='FILE PATH',
              envvar=PywbemServer.connections_file_envvar,
              help=u'File path of a YAML file containing named connection '
                   u'definitions. The default if this option is not specified '
                   u'is the file name "{df}" in the users home directory. '
                   u'EnvVar ({ev})'.
                   format(df=DEFAULT_CONNECTIONS_FILE,
                          ev=PywbemServer.connections_file_envvar))
@click.option('--pdb', is_flag=True,
              # defaulted in code
              envvar=PywbemServer.pdb_envvar,
              help=u'Pause execution in the built-in pdb debugger just before '
                   u'executing the command within pywbemcli. '
                   u'Default: EnvVar {ev}, or false.'.
                   format(ev=PywbemServer.pdb_envvar))
@click.version_option(
    message='%(prog)s, version %(version)s\npywbem, version {}'.format(
        pywbem_version),
    help=u'Show the version of this command and the pywbem package.')
@add_options(help_option)
@click.pass_context
def cli(ctx, server, connection_name, default_namespace, user, password,
        timeout, verify, certfile, keyfile, ca_certs, output_format, use_pull,
        pull_max_cnt, mock_server, verbose=None, connections_file=None,
        timestats=None, log=None, pdb=None, deprecation_warnings=None):
    """
    Pywbemcli is a command line WBEM client that uses the DMTF CIM-XML protocol
    to communicate with WBEM servers. Pywbemcli can:

    \b
    * Manage the information in WBEM servers CIM objects using the
      operations defined in the DMTF specification.  It can manage CIM
      classes, CIM instances and CIM qualifier declarations in the WBEM
      Server and execute CIM methods and queries on the server.

    \b
    * Inspect WBEM server characteristics including namespaces, registered
      profiles, and other server information.

    \b
    * Capture detailed information on communication with the WBEM
      server including time statistics and logs of the operations.

    \b
    * Maintain a persistent list of named connections to WBEM servers
      and execute operations on them by name.

    Pywbemcli implements command groups and commands to execute the CIM-XML
    operations defined by the DMTF CIM Operations Over HTTP specification
    (DSP0200).

    The general options shown below can also be specified on any of the
    commands, positioned right after the 'pywbemcli' command name.

    The width of help texts of this command can be set with the
    PYWBEMCLI_TERMWIDTH environment variable.

    For more detailed documentation, see:

        https://pywbemtools.readthedocs.io/en/stable/
    """

    # List of options that are not allowed in some cases:
    # i.e. When -name is used and when in interactive mode.
    resolved_mock_server = []
    conditional_options = ((default_namespace, 'default_namespace'),
                           (timeout, 'timeout'),
                           (verify, 'verify'),
                           (user, 'user'),
                           (certfile, 'certfile'),
                           (keyfile, 'keyfile'),
                           (ca_certs, 'ca_certs'),
                           (server, 'server'),
                           # Special because we remove some items from the
                           # list of mocks that define the mock repository
                           (resolved_mock_server, 'mock-server'))

    def create_server_instance(connection_name):
        """
        Create a PywbemServer instance from the cli arguments and return
        that object.  This method depends on using variables from the
        enclosing function. If the arguments to create a server do not exist
        no server is created.
        NOTE: connection_name is passed to this function because the
        ClickException treats it as a reference before assignment if using the
        connection_name from the enclosing scopt
        """
        # test for conflicting server definitions.
        if server or resolved_mock_server:
            if connection_name:
                click.ClickException('Option conflict: --name "{}" '
                                     'conflicts with existence of --server and '
                                     '--mock-server'.format(connection_name))
            connection_name = 'not-saved'
            pywbem_server = PywbemServer(server,
                                         resolved_default_namespace,
                                         name=connection_name,
                                         user=user,
                                         password=password,
                                         timeout=resolved_timeout,
                                         use_pull=resolved_use_pull,
                                         verify=resolved_verify,
                                         certfile=certfile,
                                         keyfile=keyfile,
                                         ca_certs=resolved_ca_certs,
                                         mock_server=resolved_mock_server)
        else:  # Server and mock_server were not specified
            # If name cmd line option, get get name from the connections
            # repo.
            if connection_name:
                validate_connections_file(connections_repo)
                if connection_name not in connections_repo:
                    raise click.ClickException(
                        "Connection definition '{}' not found in "
                        "connections file '{}'".
                        format(connection_name,
                               connections_repo.connections_file))

            # No connection name specified, try default connection name
            else:
                # get the default_connection definition
                if connections_repo.file_exists():
                    connection_name = \
                        connections_repo.get_default_connection_name()

                    # Abort because file modified to clear the bad default.
                    # This should never happen andif it does, Abort is logical.
                    if connection_name and connection_name not in \
                            connections_repo:
                        connections_repo.set_default_connection(None)
                        click.echo(
                            "Default connection definition '{}' not found in "
                            "connections file '{}'; deleted default connection "
                            "in connections file".
                            format(connection_name,
                                   connections_repo.connections_file),
                            err=True)
                        raise click.Abort()
                    if verbose:
                        click.echo('Current connection: "{}"'
                                   .format(connection_name))

            # Get the named connection from the repo
            if connection_name:
                pywbem_server = connections_repo[connection_name]
                # Test for invalid options with the --name option
                for option in conditional_options:
                    if option[0]:
                        raise click.ClickException(
                            '"--{} {}" option invalid when --name exists or '
                            'default name set.'.format(option[1], option[0]))

            else:
                # If no server defined, set None to allow commands that
                # do not require a server to be executed with no server
                # defined.
                pywbem_server = None
        return pywbem_server

    # Process cli options to produce resolved options, i.e. the
    # options with any defaults applied for non None options.
    # Produces new variables resolved... so that later tests can confirm that
    # original variables were None or not None
    pywbem_server = None
    resolved_default_namespace = default_namespace or DEFAULT_NAMESPACE
    resolved_timestats = timestats or DEFAULT_TIMESTATS
    resolved_verify = DEFAULT_VERIFY if verify is None else verify

    # There is no default ca_certs
    resolved_ca_certs = ca_certs  # None should be passed on

    # Create the connections repository object that will be included in the
    # context_object. Always use the original connections_file information
    if ctx.obj is None:
        connections_repo = ConnectionRepository(
            connections_file or DEFAULT_CONNECTIONS_PATH)
    else:
        if connections_file:
            click.ClickException("--connections-file not allowed in "
                                 "interactive mode.")
        connections_repo = ctx.obj.connections_repo

    if verbose:
        click.echo(str(connections_repo))

    if server and connection_name:
        raise click.ClickException(
            'Conflicting server definitions: name: {}, server: {}'.
            format(connection_name, server))

    if keyfile and not certfile:
        raise click.ClickException(
            'The --keyfile option "{}" is allowed only if the --certfile '
            'option is also used'.format(keyfile))

    # process mock_server option
    if mock_server:
        assert isinstance(mock_server, tuple)
        # resolve relative and absolute paths
        mock_server_path = []
        for fn in mock_server:
            if fn == os.path.basename(fn):
                mock_server_path.append(os.path.join(os.getcwd(), fn))
            else:
                mock_server_path.append(fn)

        # Test for non-existent mock files
        # TODO: Future: Create common method with code in build_respository
        for file_path in mock_server_path:
            ext = os.path.splitext(file_path)[1]
            if ext not in ['.py', '.mof']:
                raise click.ClickException(
                    "Mock file '{}' has invalid suffix '{}' "
                    "- must be '.py' or '.mof'".format(file_path, ext))
            if not os.path.isfile(file_path):
                raise click.ClickException(
                    "Mock file '{}' does not exist".format(file_path))
            if ext != '.py':
                resolved_mock_server.append(file_path)
                continue

            # The following allows executing selected python scripts at
            # startup but with only VERBOSE as a known global.  Inserted
            # primarily to support testing. If the flag exists in the
            # file, it executes immediatly but does not go into mock_server
            # variable
            with open(file_path) as fp:
                if '!PROCESS!AT!STARTUP!' in fp.readline():
                    file_source = fp.read()
                    # Only verbose is allowed here
                    globalparams = {'VERBOSE': verbose}
                    try:
                        # Using compile+exec instead of just exec allows
                        # specifying the file name, causing it to appear in
                        # any tracebacks.
                        file_code = compile(file_source, file_path, 'exec')
                        # pylint: disable=exec-used
                        exec(file_code, globalparams, None)
                    except Exception:
                        exc_type, exc_value, exc_traceback = sys.exc_info()
                        tb = traceback.format_exception(exc_type, exc_value,
                                                        exc_traceback)
                        click.echo(
                            "Mock Python process-at-startup script '{}' "
                            "failed:\n{}".format(file_path, "\n".join(tb)),
                            err=True)
                        raise click.Abort()
                else:  # not processed during startup
                    resolved_mock_server.append(file_path)

    # Simultaneous mock_server and server fails
    if server and resolved_mock_server:
        raise click.ClickException(
            'Conflicting server definitions: server: {}, mock-server: {}'.
            format(server, ', '.join(resolved_mock_server)))

    # Simultaneous mock_server and connection_name: Generate Exception.
    # Note: resolve mock_server does not include any mocks processedatstartup
    if resolved_mock_server and connection_name:
        raise click.ClickException(
            'Conflicting server definitions: mock-server: {}, name: {}'.
            format(', '.join(resolved_mock_server), connection_name))

    # Set pull default if necessary and create the resolved variable
    use_pull = use_pull or DEFAULT_PULL_CHOICE
    resolved_use_pull = USE_PULL_CHOICE[use_pull]

    resolved_pull_max_cnt = pull_max_cnt or DEFAULT_MAXPULLCNT

    resolved_timeout = timeout or DEFAULT_CONNECTION_TIMEOUT

    # Command mode (ctx is None). Processes command on comand line and quits
    # Apply the documented option defaults to create a pywbem_server instance
    # and a ContextObj instance
    if ctx.obj is None:  # No context. This is cmd line mode
        # Create the PywbemServer object (this contains all of the info
        # for the connection defined by the cmd line input)
        pywbem_server = create_server_instance(connection_name)

    # Interactive mode cmd line processing (ctx not None)
    # In interactive mode, general options specified in cmd line are used
    # to modify the pywbem_server object and the general options
    # for a single command execution.
    # This requires being able to determine for each option whether it has been
    # specified and is why general options don't define defaults in the
    # decorators that define them.
    else:  # ctx.obj exists. Processing an interactive command.
        # Apply the option defaults from the command line options
        # or from the context object.

        # If --name option, get server from connection
        if connection_name:
            validate_connections_file(connections_repo)
            try:
                pywbem_server = connections_repo[connection_name]
            except KeyError:
                raise click.ClickException(
                    "Connection definition '{}' not found in "
                    "connections file '{}'".
                    format(connection_name, connections_repo.connections_file))

        # If other parameters, modify the existing connection and reset it.
        # TODO: refactor this code to avoid deepcopy when not needed.
        else:
            if pywbem_server is None:
                # Copy to keep the original clean from any changes
                pywbem_server = deepcopy(ctx.obj.pywbem_server)
            if pywbem_server:
                modified_server = False
                if server:
                    if pywbem_server.mock_server:
                        pywbem_server.mock_server = []
                    pywbem_server.server = server
                    modified_server = True
                if mock_server:
                    if pywbem_server.server:
                        pywbem_server.server = None
                    pywbem_server.mock_server = resolved_mock_server
                    modified_server = True
                if user:
                    pywbem_server.user = user
                    modified_server = True
                if password:
                    pywbem_server.password = password
                    modified_server = True
                if verify is not None:
                    pywbem_server.verify = resolved_verify
                    modified_server = True
                if ca_certs:
                    pywbem_server.ca_certs = ca_certs
                    modified_server = True
                if certfile:
                    pywbem_server.certfile = certfile
                    modified_server = True
                if keyfile:
                    pywbem_server.keyfile = keyfile
                    modified_server = True
                if timeout:
                    pywbem_server.timeout = resolved_timeout
                if use_pull:
                    pywbem_server.use_pull = resolved_use_pull
                if server:
                    pywbem_server.server = server
                    modified_server = True
                if default_namespace:
                    pywbem_server.default_namespace = resolved_default_namespace
                    modified_server = True
                if modified_server:
                    pywbem_server.reset()
                else:
                    pywbem_server = ctx.obj.pywbem_server
            else:
                pywbem_server = create_server_instance(connection_name)

        # The following variables are maintained only in the context_obj and
        # not attached to any particular connection. If the cli argument is
        # None, this argument was not defined as part of this interactive
        # command
        if output_format is None:
            output_format = ctx.obj.output_format
        if use_pull is None:
            resolved_use_pull = ctx.obj.use_pull
        if pull_max_cnt is None:
            resolved_pull_max_cnt = ctx.obj.pull_max_cnt
        if not timestats:  # Defaults to False, not None
            resolved_timestats = ctx.obj.timestats
        if log is None:
            log = ctx.obj.log
        if verbose is None:
            verbose = ctx.obj.verbose
        if connections_repo is None:
            connections_repo = ctx.obj.connections_repo
        if pdb is None:
            pdb = ctx.obj.pdb
        if deprecation_warnings is None:
            deprecation_warnings = ctx.obj.deprecation_warnings

    # Create a command context for each command: An interactive command has
    # its own command context as a child of the command context for the
    # command line.
    ctx.obj = ContextObj(pywbem_server, output_format,
                         resolved_use_pull,
                         resolved_pull_max_cnt,
                         resolved_timestats,
                         log, verbose, pdb,
                         deprecation_warnings,
                         connections_repo)
    if verbose and os.getenv('PYWBEMCLI_DIAGNOSTICS'):
        print('CONTEXT_OBJ {!r}'.format(ctx.obj))
        print('CLICK CTX {}'.format(ctx))
        display_click_context(ctx, msg="After adding Context",
                              display_attrs=True)

    _python_nm = sys.version_info[0:2]
    if _python_nm in ((2, 7), (3, 4)):
        deprecation_warning(
            "Deprecation: Pywbemcli support for Python {}.{} is deprecated "
            "and will be removed in a future version".
            format(_python_nm[0], _python_nm[1]), ctx.obj)

    # Invoke command if one exists.
    if ctx.invoked_subcommand is None:
        ctx.invoke(repl)


@cli.command('help', options_metavar=GENERAL_OPTS_TXT)
@add_options(help_option)
@click.pass_context
def repl_help(ctx):  # pylint: disable=unused-argument
    """
    Show help message for interactive mode.
    """
    click.echo("""
The following can be entered in interactive mode:

  COMMAND                     Execute pywbemcli command COMMAND.
  !SHELL-CMD                  Execute shell command SHELL-CMD.
  <CTRL-D>, :q, :quit, :exit  Exit interactive mode.
  <CTRL-r>  <search string>   To search the  command history file.
                              Can be used with <UP>, <DOWN>
                              to display commands that match the search string.
                              Editing the search string updates the search.
  <TAB>                       Tab completion (can be used anywhere).
  -h, --help                  Show pywbemcli general help message, including a
                              list of pywbemcli commands.
  COMMAND --help              Show help message for pywbemcli command COMMAND.
  help                        Show this help message.
  :?, :h, :help               Show help message about interactive mode.
  <UP>, <DOWN>                Scroll through pwbemcli command history.

  COMMAND: May be two words (class enumerate) for commands that are within
  a group or a single word for special commands like `repl` that are not in
  a group.
""")


@cli.command('repl', options_metavar=GENERAL_OPTS_TXT)
@add_options(help_option)
@click.pass_context
def repl(ctx):
    """
    Enter interactive mode (default).

    Enter the interactive mode where pywbemcli commands can be entered
    interactively. The prompt is changed to 'pywbemcli>'.

    Command history is supported. The command history is stored in a file
    ~/.pywbemcli_history.

    Pywbemcli may be terminated from this mode by entering
    <CTRL-D>, :q, :quit, :exit

    In the repl mode, <CTRL-r> man be used to initiate an interactive
    search of the history file.

    Interactive mode also includes an autosuggest feature that makes
    suggestions from the command history as the command the user types in the
    command and options.
    """

    history_file = PYWBEMCLI_HISTORY_FILE
    if history_file.startswith('~'):
        history_file = os.path.expanduser(history_file)

    click.echo("Enter 'help' for help, <CTRL-D> or ':q' "
               "to exit pywbemcli or <CTRL-r> to search history, ")

    prompt_kwargs = {
        'message': PYWBEMCLI_PROMPT,
        'history': FileHistory(history_file),
    }

    if USE_AUTOSUGGEST:
        prompt_kwargs['auto_suggest'] = AutoSuggestFromHistory()

    click_repl.repl(ctx, prompt_kwargs=prompt_kwargs)
